import sys
from .MMlib3 import nogap, printerr, write, symmetrical_dict
try:  
  from PyQt5 import QtCore, QtGui, QtWidgets
  from ete3 import faces
except: 
  printerr('ERROR PyQt5 and ete3 must be installed to use this program!', 1)
  raise 


def distance_matrix_from_tree(root):
  """ Given a tree structures, returns a  symmetrical_dict  object with distance from every leaf to every other.
  To parse it:  
  dm=distance_matrix_from_tree(root)
  for n1 in dm.all_keys():
    for n2 in dm.all_keys():
      if n1!=n2:
        print n1.name, n2.name, dm[n1][n2]  
  """
  dmat=symmetrical_dict()
  for node in root.traverse(strategy='postorder'):
    if node.is_leaf(): node._dist_leaves={ node : node.dist }
    else:
      node._dist_leaves={}
      for child1_index, child1 in enumerate(node.children):  # in binary trees:  only one child1, one child2 cycle
        for leaf1 in child1._dist_leaves:
          dist_leaf1=child1._dist_leaves[leaf1]
          node._dist_leaves[leaf1]=dist_leaf1 + node.dist   
          for child2 in node.children[child1_index+1:]: 
            for leaf2 in child2._dist_leaves:
              dist_leaf2=child2._dist_leaves[leaf2]
              #assert( not dmat.has_keys(leaf1, leaf2) )
              dmat[leaf1][leaf2]=dist_leaf1+dist_leaf2
              
  return dmat

#def detect_contaminations_from_gene_tree(tree, species_fn, taxonomy_fn):
def get_taxonomy_consistency_score(tree, species_fn, taxonomy_fn, neighbours=10, distance_matrix=None):
  """ Given a gene tree;  
  and a species_fn   to derive species given a node;  
  and a taxonomy_fn  to derive a tax_string from species (tax_string such as 
"cellular organisms; Eukaryota; Opisthokonta; Metazoa; Eumetazoa; Bilateria; Protostomia; Ecdysozoa; Panarthropoda; Arthropoda; Chelicerata; Arachnida; Araneae; Araneomorphae; Entelegynae; Orbiculariae; Araneoidea; Theridiidae; Latrodectus"
  return a dictionary   {node : taxonomy_consistency_score, normalized_score_by_neighbours} 
   computed for each leaf on its n={neighbours} closest leaves
 distance_matrix can be provided, otherwise it is computed
 """
  sp2tax_list={} #caching this
  if distance_matrix is None:        distance_matrix=distance_matrix_from_tree(tree)
  node2av_cons_score={}
  node2neighbours={}
  for leaf in tree:
    leaf_sp=species_fn(leaf)
    if not leaf_sp in sp2tax_list:     sp2tax_list[leaf_sp]=taxonomy_fn(leaf_sp).split('; ')
    leaf_tax=sp2tax_list[leaf_sp]  ## this is a list

    closest_neighbours=sorted([leaf2   for leaf2 in tree if leaf!=leaf2],   key=lambda x:distance_matrix[leaf][x]) [:neighbours]
    node2neighbours[leaf]=closest_neighbours

    cons_score_list=[]
    for leaf2 in closest_neighbours:
      leaf2_sp=species_fn(leaf2)
      if not leaf2_sp in sp2tax_list:     sp2tax_list[leaf2_sp]=taxonomy_fn(leaf2_sp).split('; ')
      leaf2_tax=sp2tax_list[leaf2_sp]  ## this is a list
      matches=0
      for ibit, bit1 in enumerate(leaf_tax):
        if ibit == len(leaf2_tax): break
        bit2=leaf2_tax[ibit]
        if bit1==bit2:         matches+=1
        else:                  break
      cons_score=matches/float(len(leaf_tax))
      cons_score_list.append(cons_score)
    av_cons_score=sum(cons_score_list)/len(cons_score_list)
    node2av_cons_score[leaf]=[av_cons_score]
  for leaf in tree:
    closest_neighbours=node2neighbours[leaf]
    av_cons_scores= [node2av_cons_score[leaf2][0]   for leaf2 in  closest_neighbours]
    normalized_score_by_neighbours=   node2av_cons_score[leaf][0] / ( sum(av_cons_scores)   /   len(av_cons_scores) )
    node2av_cons_score[leaf].append(normalized_score_by_neighbours)
  return node2av_cons_score
              
def get_feature_consistency_score(tree,  feature_fn, neighbours=10, distance_matrix=None, bigger_ok=False):
  """Generalization of get_taxonomy_consistency_score in which we look at any numerical feature, and compare it across neighbours to detect nodes with outlier features
     Prototyping this with sequence length.
     
"""
  node2feat={} #caching this
  if distance_matrix is None:        distance_matrix=distance_matrix_from_tree(tree)
  node2cons_score={}
  node2neighbours={}
  for leaf in tree:
    #leaf_sp=species_fn(leaf)
    #if not leaf_sp in sp2tax_list:     sp2tax_list[leaf_sp]=taxonomy_fn(leaf_sp).split('; ')
    if not leaf in node2feat:    node2feat[leaf]=feature_fn(leaf)
    leaf_feat=node2feat[leaf]
    #leaf_tax=sp2tax_list[leaf_sp]  ## this is a list

    closest_neighbours=sorted([leaf2   for leaf2 in tree if leaf!=leaf2],   key=lambda x:distance_matrix[leaf][x]) [:neighbours]
    node2neighbours[leaf]=closest_neighbours

    #cons_score_list=[]
    feat_neighbours=[]
    for leaf2 in closest_neighbours:
      if not leaf2 in node2feat:    node2feat[leaf2]=feature_fn(leaf2)
      leaf2_feat=node2feat[leaf2]
      ### compute proportion between the smallest and the largest of the two. That's our score.
      feat_neighbours.append(leaf2_feat)

    av_neighbours=sum(feat_neighbours)/float(len(feat_neighbours))
    if leaf_feat > av_neighbours:
      if bigger_ok: cons_score=1.0
      else:         cons_score=  av_neighbours/leaf_feat
    else:           cons_score=  leaf_feat/av_neighbours
    node2cons_score[leaf]=[cons_score]
  for leaf in tree:
    closest_neighbours=node2neighbours[leaf]
    cons_scores= [node2cons_score[leaf2][0]   for leaf2 in  closest_neighbours]
    normalized_score_by_neighbours=   node2cons_score[leaf][0] / ( sum(cons_scores)   /   len(cons_scores) )
    node2cons_score[leaf].append(normalized_score_by_neighbours)
  return node2cons_score      
  

def make_clusters(root, species_fn=None, min_species_dup=10, min_div_dist=0.35, min_species_div=10):
  """ Given a PhyloTree with evol_events, automatically calls clusters of genes. Two methods:
-1 Duplication: The node is a duplication node with large support (n_species >= min_species_dup )
-2 Divergence:  The node has a large branch  (dist>=min_div_dist), and contains many species (n_species >= min_species_div)
  Returns a list and a dict:  [ (node, clusterid) ], { leaf: clusterid }     first list is ordered like traverse preorder (root to leaf)
   In the dict leaves are assigned to the cluster of the closest parent node assigned to a cluster.
   """
  def get_species(node):
    if species_fn is None: return node.species 
    return species_fn(node)

  def get_species_set(node):
    if node.is_leaf():      return set([get_species(node)])
    elif hasattr(node, '_species_set'):      return node._species_set
    else:
      node._species_set=set()
      for n in node.children:  
        node._species_set= node._species_set.union(  get_species_set(n)  )
      return node._species_set

  node2clustername={}
  sorted_nodes=[]
  leaf2clustername={}
  clusternames=set()
  for node in root.traverse():
    if node.is_leaf(): continue
    is_cluster=False
    # if not  hasattr(node, 'evoltype'):
    #   write('ERR with node: {n}'.format(n=node), 1, how='red')
    #   continue
    if node.evoltype == 'D':
      a, b = node.children
      set_a =get_species_set(a)
      set_b =get_species_set(b)
      n_species= len(set_a.intersection(set_b))
      if n_species>=min_species_dup:       
        is_cluster=True
    if not is_cluster:
      if node.dist>= min_div_dist and len([n for n in node])>=min_species_div:
        is_cluster=True
    if is_cluster:
      clustername_root=None
      parent=node.up 
      while parent: 
        if parent in node2clustername: clustername_root=node2clustername[parent]; break
        parent=parent.up 
      if clustername_root is None: clustername_root='c'
      index=0
      while '{r}{i}'.format(r=clustername_root, i=index) in clusternames: index+=1
      clustername='{r}{i}'.format(r=clustername_root, i=index)
      node2clustername[node]=clustername
      sorted_nodes.append(node)
      for leaf in node:          leaf2clustername[leaf]=clustername
      clusternames.add(clustername)
      #write(clustername, 1, how='magenta')
      #write(node, 1)
  for n in root.traverse():
    if hasattr(node, '_species_set'):      del node._species_set
  return  [ (n, node2clustername[n]) for n in sorted_nodes], leaf2clustername



global warnings_reduced_text; warnings_reduced_text={} #keys: categories
def shrink_font_to_box (simpleTextItem, width=-1, height=-1, category='unknown'):
  """Function to shrink a piece of text in a rectangle of a given width and height; its initial font size value is reduced until it fits. The global variable warning_reduced_text keeps track of every time this function is run succesfully. Note: after calling this function you set the positions of the simpleTextItem object according to its renewed boundingRect """
  psize=99 #not used, this is just for the first time last condition of while is checked
  reducing_of_points=0
  while (width>0 and simpleTextItem.boundingRect().width() > width) or (height>0 and simpleTextItem.boundingRect().height()> height ) and psize>1:
    font=simpleTextItem.font()
    psize= font.pointSize()
    font.setPointSize(psize-1)
    simpleTextItem.setFont(font)
    reducing_of_points+=1
  if reducing_of_points:
    if category not in warnings_reduced_text: warnings_reduced_text[category]=0
    warnings_reduced_text[category]+=1

class GapFace(faces.Face):
  """ Simple face that can be used as place holder, e.g. to indicate gaps in the syntheny view. A single horizontal dotted line is drawn at the center. Width and height can be used so that this can be aligned to other faces.""" 
  def __init__(self, width=150, height=40, color='gray', pen_size=1):
    faces.Face.__init__(self)
    self.type = "item"
    self.item = None
    self.width=width
    self.height=height
    self.color=color
    self.pen_size=pen_size
  def _width(self):        return self.item.rect().width()
  def _height(self):       return self.item.rect().height()
  def update_items(self):
      self.item = QtWidgets.QGraphicsRectItem(0, 0, self.width, self.height) # backbone object, no color 
      self.item.setPen(QtGui.QPen(QtCore.Qt.NoPen)) #no black border      
      if not self.color is None:
        outline_pen = QtGui.QPen(    QtGui.QColor(self.color),  self.pen_size, QtCore.Qt.DotLine)
        gap_line =  QtWidgets.QGraphicsLineItem( 0, self.height/2.0,  self.width,  self.height/2.0,            self.item)
        gap_line.setPen(outline_pen)

class ContigTerminusFace(faces.Face):
  """ Simple face that can be used as place holder, to indicate the end of this contig (or the start)  in the syntheny view. Width and height can be used so that this can be aligned to other faces. Init as: cef= ContigEndFace(width=100, height=40, thickness=5, arm_width=20, arm_height=30, arcolor='gray', mode='start')    OR: mode='end'
  <              width           >
  ################################ ^            mode=start                   mode=end
  #       <        > arm_width   #                                    #   
  #     ^ ---------| ^           # h        x1,y4             x3,y4   #   x3,y4               x1,y4
  #       | -------| v thickness # e              x2,y3       x3,y3   #   x3,y3         x2,y3
  # arm_  | |                    # i                                  #
  # height| |                    # g                                  #
  #       | -------|             # h              x2,y2       x3,y2   #   x3,y2         x2,y2
  #     v ---------|             # t        x1,y1             x3,y1   #   x3,y1               x1,y1
  #       < > thickness          #                                    #
  ################################ v
""" 
  def __init__(self, width=150, height=40, thickness=5, arm_width=20, arm_height=30,color='darkgray', mode='start'):
    faces.Face.__init__(self)
    self.type = "item"
    self.item = None
    if arm_width  > width :  arm_width=  width
    if arm_height > height : arm_height= height
    thickness= min([thickness, width, height])
    self.width=width;     self.height=height; 
    self.color=color; self.mode= mode
    self.thickness=thickness;   self.arm_width=arm_width;  self.arm_height=arm_height

  def _width(self):        return self.item.rect().width()
  def _height(self):       return self.item.rect().height()
  def update_items(self):
      self.item = QtWidgets.QGraphicsRectItem(0, 0, self.width, self.height) # backbone object, no color 
      self.item.setPen(QtGui.QPen(QtCore.Qt.NoPen)) #no black border      
      fill_brush=  QtGui.QBrush(  QtGui.QColor(self.color)  )
      #pen_size=1
      #outline_pen = QtGui.QPen(    QtGui.QColor(self.color),  pen_size, QtCore.Qt.DotLine)
      if self.mode=='start':        x1 = (self.width - self.arm_width) / 2.0;       x2 = x1 + self.thickness;       x3 = x1 + self.arm_width
      elif self.mode=='end':        x3 = (self.width - self.arm_width) / 2.0;       x1 = x3 + self.arm_width;       x2 = x1 - self.thickness
      y4 = (self.height - self.arm_height) / 2.0;     y3 = y4 + self.thickness;       y1 = y4 + self.arm_height;      y2 = y1 - self.thickness
      qpoints_list= [    #adding the seven points to contruct the arrow. starting from top left corner (when start mode)
                 QtCore.QPointF( x1,y4 ),  QtCore.QPointF( x3,y4 ),   QtCore.QPointF( x3,y3 ),   QtCore.QPointF( x2,y3 ),   \
                 QtCore.QPointF( x2,y2 ),  QtCore.QPointF( x3,y2 ),  QtCore.QPointF( x3,y1 ),   QtCore.QPointF( x1,y1 )     ]
      ear=   QtGui.QGraphicsPolygonItem(   QtGui.QPolygonF(  qpoints_list ),   self.item)  
      ear.setPen(  QtGui.QPen(QtCore.Qt.NoPen)  )
      ear.setBrush(fill_brush)      
      
class GeneFace(faces.Face):
  """Mother class for some Face representations created for genes. It is not functional by itself, it requires some functions (e.g. update_items) to be defined in the subclasses.
     Initialized as:
     gf= GeneFace(gene_object,   **key_attributes )
     The gene object is of class MMlib.gene; in its minimal form, it must contain a ".strand" attribute (+ or -), a .chromosome attribute, and an .id attribute. By default id and chromosome are printed within the representation, but this can be controlled with the following keyattribute: 

    *printed: a hash with as keys, some keywords of attributes of the self.gene object that will be printed as text.
      - possible values: ['id', 'chromosome', 'boundaries', 'text']  OR any other attribute found in self.gene; the vertical area will be partitioned in an equal number of horizontal stripes to accomodate all the information, with order defined by attribute order_text (any additional attribute, not included in order_text, goes as last). 'id', 'chromosome' and 'text' are looked for in the self.gene object as text attributes (note: if self.gene.text is not defined, an error occurs). For boundaries, there's an hardcoded routine that looks in the self.gene.exon coordinates and print a summary of it, using an text separator the self.gene.strand. Example: strand:+ positions:3-30,500-785   --> summary is 3+785. 
    *colors: this hash defines the colors for all objects displayed. As keys, you may provide annotation names (for subclass ExonViewFace) or any of these reserved keywords:
      - bkg:       the fill color of the main gene rectangle
      - outline:   the line color around the main gene rectangle
      - gap:       the color of the line denoting gaps  (which in dotted style)  #only for ExonViewFace
      - text:      the color of all text displayed 
"""
  default_attributes={
      'order_text':['id', 'chromosome', 'boundaries', 'text'],
      'height':40, 'width':150,  #'width_per_position':None, 
      'font':'Courier New', 'font_size':12, 'shrink_font_to_fit':True, 'font_margin':4,
      'colors':{'bkg':'#4444AA', 'outline':'#000000', 'text':'#000000' },
      'printed':{'id':1, 'chromosome':1, 'text':0, 'boundaries':0},
      'pen_size':1,
      'rotable':False,
    }

  def _width(self):        return self.item.rect().width()
  def _height(self):       return self.item.rect().height()

  def __init__(self, gene_obj, **attributes):
      faces.Face.__init__(self)
      self.type = "item"
      self.item = None
      self.gene= gene_obj
      new_attribute_keys ={}   # to keep track of those which are not defined in the default dict;  key: attribute name --> value: list of keys
      for attribute_name in attributes: 
        if attribute_name in self.default_attributes:
          if type( attributes[attribute_name] ) == dict:
            for k in list(attributes[attribute_name].keys()):  #### just checking
              if k not in self.default_attributes[attribute_name]:
                if  not (attribute_name in ['colors', 'printed']): printerr('init GeneFace WARNING  the key "'+str(k)+'" of attribute "'+attribute_name+'" is not present for this class!', 1)
                else:
                  if attribute_name not in new_attribute_keys: new_attribute_keys[attribute_name]=[]
                  new_attribute_keys[attribute_name].append(k)                         
        else:        printerr('init GeneFace WARNING  the attribute "'+attribute_name+'" is not present for this class!', 1 )
      ### parsing attributes: set
      for attribute_name in self.default_attributes:
        if attribute_name in attributes:
          if type(self.default_attributes[attribute_name]) != dict:   setattr(self, attribute_name, attributes[attribute_name])
          else: 
            h={}
            keys_to_parse=list(self.default_attributes[attribute_name].keys())
            if attribute_name in new_attribute_keys:  keys_to_parse += new_attribute_keys[attribute_name]
            for k in keys_to_parse:
              if  k in attributes[attribute_name]:    h[k]=attributes[attribute_name][k]
              else:                                         h[k]=self.default_attributes[attribute_name][k]  
            setattr(self, attribute_name, h)            
        else:          setattr(self, attribute_name, self.default_attributes[attribute_name])  # value not defined in __init__

  def sort_fields_to_print(self, field_list):
      """ Used to determine the vertical order of things that must be printed (self.text attribute)""" 
      k_hash={}
      for field in field_list:
        position=None
        #finding index of field in self.order_text
        for index, field_name in enumerate(self.order_text): 
          if field==field_name: 
            position=index
            break
        if position is None: position=len(self.order_text)
        k_hash[field]=position
      return sorted(field_list, key=lambda x:k_hash[x])

  def update_items(self):        raise Exception("ERROR calling the raw update_items function from class GeneView! This function must be defined in the subclasses. Aborting! ")

  def draw_text_fields(self):
      ### printing any text 
      fields_to_print=self.sort_fields_to_print( [ f for f in self.printed if self.printed[f] ] )  # margin #--area-field1--# margin #--area_field2--# margin ... #--area_fieldN--# margin #
      if fields_to_print:
        area_field_height=( self.height - self.font_margin*(len(fields_to_print)+1) ) / float( len(fields_to_print) )
        font_used=  QtGui.QFont( self.font, self.font_size )    
        brush_used= QtGui.QColor( self.colors['text']   )
        for field_index, field_name in enumerate( fields_to_print ):
          text_obj = QtGui.QGraphicsSimpleTextItem()
          text_obj.setBrush( brush_used  )
          text_obj.setFont( font_used )
          if field_name.startswith('&'):      the_text = str(  getattr(self.gene, field_name[1:]) () )  #it's a function; written like &function_name
          elif field_name == 'boundaries':    
            join_symbol=self.gene.strand
            the_text =   join_symbol.join(list(map(str, self.gene.boundaries())))
          else:                     the_text = str(  getattr(self.gene, field_name)  )  #generic for attribute
          text_obj.setText(the_text)
          if self.shrink_font_to_fit:  
            shrink_font_to_box(text_obj, self.width, area_field_height, category=field_name)  #reducing font if necessary to fit the allocated space
      ## move position to center it
          br=text_obj.boundingRect();       x_text= ( self.width - br.width()) /2;    
          y_text= self.font_margin +   (area_field_height+self.font_margin)*field_index   +   ( area_field_height - br.height() )/2    
          text_obj.setPos( x_text, y_text )
          text_obj.setParentItem(self.item)

class GeneViewFace(GeneFace):
  """ Typical GFF visualization, representing exons and introns. Something like this:
    _______     ________
 --|       |---|        |--     
    -------     --------
Usage:     gvf= GeneViewFace(gene_object, introns=[], xmin=None, xmax=None, shrink_introns=None, **key_attributes )
     The gene object is of class MMlib.gene; in its minimal form, it must contain a ".strand" attribute (+ or -), a .chromosome attribute, and an .id attribute. 
    *introns can be used to display connectors in the typical triangle style. These can link positions which are not necessarily exon boundaries.  Argument must be provided in the form: introns=[ [start1,stop1], [start2, stop2] .. ], where start < stop
    *xmin and *xmax define the boundaries of the genomic region displayed. If None, the min and max values present in the gene exons or in the introns argument are used
    *shrink_introns: often introns are so much larger than exons that it is bad for visualization. Set this to any integer number to set the maximum intron size which is faithfully represented; introns bigger than this will be displayed as big as this, with a pattern indicating the compression.
     By default the .id attribute of the gene object (and only that) is printed within the representation, but this can be controlled with the following keyattribute: 
    *printed: a hash with as keys, some keywords of attributes of the self.gene object that will be printed as text.
      - possible values: ['id', 'chromosome', 'boundaries', 'text']  OR any other attribute found in self.gene; the vertical area will be partitioned in an equal number of horizontal stripes to accomodate all the information, with order defined by attribute order_text (any additional attribute, not included in order_text, goes as last). 'id', 'chromosome' and 'text' are looked for in the self.gene object as text attributes (note: if self.gene.text is not defined, an error occurs). For boundaries, there's an hardcoded routine that looks in the self.gene.exon coordinates and print a summary of it, using an text separator the self.gene.strand. Example: strand:+ positions:3-30,500-785   --> summary is 3+785. 
    *colors: this hash defines the colors for all objects displayed. As keys, you may provide annotation names (for subclass ExonViewFace) or any of these reserved keywords:
      - bkg:       the fill color of the main gene rectangle
      - outline:   the line color around the main gene rectangle
      - text:      the color of all text displayed 
      - line:      the color of the genome line
      - intron:    the color of the introns drawn
      - exonN or exonN.outline:   define a key like this, with N replacing the index of any exon (1 is first at 5'), to change the color of the background (or outline) of any specific exon. Useful to highlight an exon
    SIZE:
     for height, you must define either: *height (total height of face in pixels) or *height_exons and height_introns (to define manually the space attributed to the exon and the intron linker.
     Similarly for width you must define either *width (total width of face in pixels)  or *width_per_position, where position refers to the coordinates in the gene object
    *pen_size defines the size of the pen used to draw all lines (default: 1)
    *intron_pen_size defines the size of the pen used for introns only (default: 2)
"""
  default_attributes={
      'order_text':['id', 'chromosome', 'boundaries', 'text'],
      'height':None, 'height_exons':40,  'height_introns':15, 
      'width':150,  'width_per_position':None, 
      'font':'Courier New', 'font_size':12, 'shrink_font_to_fit':True, 'font_margin':4,
      'colors':{'bkg':'#4444AA', 'outline':'#000000', 'text':'#000000', 'line':'#000000', 'intron':'#000000', #'exon1':'#888844', 'exon1.outline':'#AA0000'
      },
      'printed':{'id':1, 'chromosome':0, 'text':0, 'boundaries':0},
      'pen_size':1,       'intron_pen_size':2,
      'rotable':False,
    }

  def __init__(self, gene_obj, introns=[], xmin=None, xmax=None,  shrink_introns=None, **attributes):
      GeneFace.__init__(self, gene_obj, **attributes)      ### attributes are added / updated here
      ## setting height
      if not introns: self.height_introns=0
      if self.height is None:    self.height = self.height_exons  +   self.height_introns  
      elif not introns:          self.height_exons = self.height
      else:             
        self.height_exons =    int(0.6 * self.height)
        self.height_introns = self.height - self.height_exons

      # to determine self.width_per_position I need to know how many introns are compressed. 
      self.compressed_regions=[]   #[  [start, end, shrinked_to], ...  ]    --> shrinked_to meaning, the apparent width in the compressed version
      self.compressed_intron_indexes={}
      compressed_to_length    = 0;         total_length_compressed = 0
      if shrink_introns:
        for intron_index, intron_positions in enumerate( self.gene.introns().exons ):
          intron_start, intron_end = intron_positions
          if intron_end - intron_start +1 > shrink_introns:
            self.compressed_regions.append(  [intron_start, intron_end, shrink_introns]  )
            self.compressed_intron_indexes[intron_index]=True
        total_length_compressed = sum([x[1]-x[0]+1 for x in self.compressed_regions])
        compressed_to_length  = shrink_introns * len(self.compressed_regions)
                                    
      ### get or set self.width and self.width_per_position   and min and max boundaries
      boundaries=self.gene.boundaries()
      if xmin is None: xmin=min( boundaries +  [i_start for i_start,i_end in introns] + [i_end for i_start,i_end in introns] )
      if xmax is None: xmax=max( boundaries +  [i_start for i_start,i_end in introns] + [i_end for i_start,i_end in introns] )
      if self.width_per_position:     self.width=             self.width_per_position *  (xmax-xmin+1 - total_length_compressed + compressed_to_length )
      elif self.width:                self.width_per_position=float(self.width) /        (xmax-xmin+1 - total_length_compressed + compressed_to_length )
      else: raise Exception("GeneViewFace ERROR neither width or width_per_position are set!")

      self.processed_coords = {'exons':[], 'introns':[]} #, 'compressed':[]}   # [ [x1, x2] ...  ]  in pixels   ##### referring to exon positions as "positions" and to pixel positions as "coords"
      for exon_index, exon_positions in enumerate(self.gene.exons):
        exon_start, exon_end= exon_positions
        if self.gene.strand=='-':                     delta_pos_start = xmax - exon_end;          delta_pos_end   = xmax - exon_start          
        else:                                         delta_pos_start = exon_start - xmin;        delta_pos_end   = exon_end   - xmin 

        ## correcting for compressed introns
        for compressed_index, compr_indexes in enumerate( self.compressed_regions ):
            compressed_start, compressed_end, shrinked_to = compr_indexes
            all_compressed_are_behind= exon_start < compressed_start
            if self.gene.strand=='-': all_compressed_are_behind = not all_compressed_are_behind
            if all_compressed_are_behind:  break
            diff_for_compression = shrinked_to - (compressed_end - compressed_start +1) 
            delta_pos_start += diff_for_compression;  delta_pos_end += diff_for_compression
                    
        coord_start = delta_pos_start * self.width_per_position;   coord_end = delta_pos_end * self.width_per_position 
        self.processed_coords['exons'].append([coord_start, coord_end])

      for intron_index, intron_positions in enumerate(introns):
        intron_start, intron_end= intron_positions
        if self.gene.strand=='-':          delta_pos_start = xmax - intron_end;         delta_pos_end   = xmax - intron_start  
        else:                              delta_pos_start = intron_start - xmin;       delta_pos_end   = intron_end   - xmin 
        ## correcting for compressed introns
        #print "******", self.gene.id,  intron_start, intron_end
        for compressed_index, compr_indexes in enumerate( self.compressed_regions ):
            compressed_start, compressed_end, shrinked_to = compr_indexes
          #  print compressed_start, compressed_end, shrinked_to
            if    self.gene.strand=='+':  all_compressed_are_behind=  intron_end   < compressed_start  # if this intron is also compressed, this becomes False, but we add it to end only 
            elif  self.gene.strand=='-':  all_compressed_are_behind=  intron_start > compressed_end

            diff_for_compression = shrinked_to - (compressed_end - compressed_start +1) 
            if all_compressed_are_behind:                break
            #else:               print "summing diff!", compressed_index, diff_for_compression
            if    (self.gene.strand=='+' and intron_start>compressed_end) or (self.gene.strand=='-' and intron_end<compressed_start):
              delta_pos_start += diff_for_compression
         #     print "diff start", diff_for_compression
         #   print "diff end", diff_for_compression
            delta_pos_end += diff_for_compression

        coord_start     = delta_pos_start * self.width_per_position;   coord_end     =  delta_pos_end * self.width_per_position
        self.processed_coords['introns'].append([coord_start, coord_end])
      self.xmin=xmin;       self.xmax=xmax; self.introns=introns;     self.shrink_introns=shrink_introns
                
  def update_items(self):
      """ the actual function doing the graphics """
      self.item = QtWidgets.QGraphicsRectItem(0, 0, self.width, self.height) # backbone object, no color 
      self.item.setPen(QtGui.QPen(QtCore.Qt.NoPen)) #no black border      

      if self.colors['outline'] is None: outline_color= self.colors['bkg']
      else: outline_color= self.colors['outline']
      outline_pen = QtGui.QPen(    QtGui.QColor(outline_color),  self.pen_size, QtCore.Qt.SolidLine)
      exon_brush  = QtGui.QBrush(  QtGui.QColor(self.colors['bkg']) )
      line_pen    = QtGui.QPen(    QtGui.QColor( self.colors['line']  ),  self.pen_size, QtCore.Qt.SolidLine)
      intron_pen =  QtGui.QPen(    QtGui.QColor(self.colors['intron']),   self.intron_pen_size, QtCore.Qt.SolidLine)

      if self.processed_coords['exons']:
        ## draw line up to first exon box
        coord_start_line = 0
        coord_end_line   = self.processed_coords['exons'][0][0]-1
        if int( coord_end_line - coord_start_line  ) > 0:
#          print 'drawing first line', coord_start_line, coord_end_line
          line =QtWidgets.QGraphicsLineItem( coord_start_line   , self.height_introns + self.height_exons / 2.0,   coord_end_line,   self.height_introns + self.height_exons / 2.0,           self.item)             
          line.setPen(line_pen);          
        ####### drawing exons
        for exon_index, exon_coords in enumerate(self.processed_coords['exons']):
          coord_start, coord_end = exon_coords
          colored_rect = QtWidgets.QGraphicsRectItem(  coord_start, self.height_introns, coord_end-coord_start+1, self.height_exons, self.item)
          ### custom color specified for this exon?
          if  'exon'+str(exon_index+1) in self.colors:   colored_rect.setBrush(QtGui.QBrush(QtGui.QColor(self.colors['exon'+str(exon_index+1)])))
          else:           colored_rect.setBrush(exon_brush)          
          if  'exon'+str(exon_index+1)+'.outline' in self.colors:   colored_rect.setPen(QtGui.QPen(QtGui.QColor(self.colors['exon'+str(exon_index+1)+'.outline']), self.pen_size, QtCore.Qt.SolidLine  ))
          else:           colored_rect.setPen(outline_pen)

          ## drawing line just after this
          if   exon_index+1 < len(  self.processed_coords['exons']  ):  ## not last exon
            coord_start_line=coord_end +1 
            coord_end_line=  self.processed_coords['exons'][exon_index+1][0] -1   #start of next exon
            #print "drawing ", coord_end+1, coord_next_exon_start-1
          else:  # last exon
            coord_start_line=coord_end +1 
            coord_end_line  = self.width
          if int( coord_end_line - coord_start_line  ) > 0:
            #print "drawing last line", coord_start_line, coord_end_line
            if not exon_index in self.compressed_intron_indexes:
              line =QtWidgets.QGraphicsLineItem( coord_start_line   , self.height_introns + self.height_exons / 2.0,   coord_end_line,   self.height_introns + self.height_exons / 2.0,      self.item);                 line.setPen(line_pen);          
            else:
              compressed_start, compressed_end = coord_start_line, coord_end_line
              size= compressed_end - compressed_start +1

        ####### drawing compressed      
        ##          p4 / /p7              y2
        ##   p1     p2/ /p6
        ##    -------/ /------- p8        y1
        ##          / /
        ##       p3/ /p5                  y0
        ##
              y0=self.height_introns + 0.9*self.height_exons 
              y1=self.height_introns + self.height_exons/2.0
              y2=self.height_introns + 0.1*self.height_exons 
              p1=compressed_start
              p2=compressed_start + 0.4 * size
              p3=compressed_start + 0.3 * size
              p4 = p5 = compressed_start + 0.5 * size
              p6=compressed_start + 0.6 * size
              p7=compressed_start + 0.7 * size
              p8= compressed_end          
              # horiz line #1
              line =QtWidgets.QGraphicsLineItem( p1, y1, p2, y1, self.item);         line.setPen(line_pen);
              # horiz line #2
              line =QtWidgets.QGraphicsLineItem( p6, y1, p8, y1, self.item);         line.setPen(line_pen);
              # rising line #1
              line =QtWidgets.QGraphicsLineItem( p3, y0, p4, y2, self.item);         line.setPen(line_pen);
              # rising line #2
              line =QtWidgets.QGraphicsLineItem( p5, y0, p7, y2, self.item);         line.setPen(line_pen);

      if self.processed_coords['introns']:        
        ####### drawing introns
        for intron_index, intron_coords in enumerate(self.processed_coords['introns']):
          coord_start, coord_end = intron_coords
          halfway= (coord_end + coord_start) / 2.0
          # rising line
          line =QtWidgets.QGraphicsLineItem( coord_start, self.height_introns , halfway, 0 , self.item)
          line.setPen(intron_pen);
          # descending line
          line =QtWidgets.QGraphicsLineItem( halfway, 0 , coord_end, self.height_introns , self.item)
          line.setPen(intron_pen);
        
      self.draw_text_fields()
      return  

class ArrowGeneFace(GeneFace):
  """ Thought for syntheny; each gene is displayed as a colored arrow, towards right or left (+ or - strand). If the gene.strand is None, it's just a rectangle (no arrow)
Initialize like:     agf= ArrowGeneFace( gene_object,  **key_attributes)
Gene object is of class MMlib.gene; these attributes must be defined: strand, chromosome, id.
Example to obtain the gene object:   gene_object=gene(strand='+', chromosome='chr1', id='Flybase0001')
Possible **key_attributes:
  -height, width: these refer to the (invisible) external box surrounding the arrow, in pixels
  -arrow_height and pointer_width ; see figure below
  -colors: a hash defining the colors to draw this. Example: colors={'bkg':'lightblue', 'outline':'#000000', 'text': 'red'}.  See GeneFace __doc__ string for more details. If outline is None, the same color as for bkg is used. If box_bkg or box_line are not provided, none (transparent) is used
  -reverse: if this is True, the arrow will point to the opposite direction as it would be otherwise
  -printed: decides what text is printed in the figure. Text is provided as attributes of the gene object. See GeneFace __doc__ string for more details. By default in this class, only the gene.id attribute is printed
  -others: rotable, font, font_size, shrink_font_to_fit, font_margin : see GeneFace

     < ---      width      --- >
  ^  ###########################
  |  #  <box_bkg>        |\    #
  |  #-------------------|  \  #   -
 hei #|      < bkg >         ) #   |--> arrow_height
 ght #-------------------|  /  #   -
  |  #      outline ^    |/    #
  -  ###########################
     box_line ^          <--->
                      pointer_width
  """
  default_attributes={
      'order_text':['id', 'chromosome', 'text', 'boundaries'],
      'height':40, 'width':150,  #'width_per_position':None, 
      'arrow_height':25,      'pointer_width':30,
      'font':'Courier New', 'font_size':8, 'shrink_font_to_fit':False, 'font_margin':4,
      'colors':{'bkg':'#4444AA', 'outline':None, 'text':'#000000', 'box_bkg':None, 'box_line':None },
      'printed':{'id':1, 'chromosome':0, 'text':0, 'boundaries':0},
      'pen_size':1, 'reverse':False, 
      'rotable':False,
    }
  
  def update_items(self):
      """ the actual function doing the graphics """
      if not self.gene.strand in ['+', '-', None]: raise Exception("ArrowGeneFace ERROR this face must be initialized with a gene object with a .strand attribute which is + or -  (or None) ! This was received: {0}".format(self.gene.strand))
      self.item = QtWidgets.QGraphicsRectItem(0, 0, self.width, self.height) # backbone object, no color 
      if not 'box_line' in self.colors or self.colors['box_line'] is None:   box_pen=QtGui.QPen(QtCore.Qt.NoPen)
      else:                                                                  box_pen=QtGui.QPen(  QtGui.QColor(self.colors['box_line']), self.pen_size, QtCore.Qt.SolidLine )
      if not 'box_bkg' in self.colors or self.colors['box_bkg'] is None:     box_brush=QtGui.QBrush(  QtCore.Qt.NoBrush  )
      else:                                                                  box_brush=QtGui.QBrush(  QtGui.QColor(self.colors['box_bkg'])  )
      self.item.setPen(box_pen) #no black border      
      self.item.setBrush(box_brush)      

      if not 'outline' in self.colors or self.colors['outline'] is None: outline_color= self.colors['bkg']
      else:           outline_color= self.colors['outline']
      outline_pen = QtGui.QPen(    QtGui.QColor(outline_color),  self.pen_size, QtCore.Qt.SolidLine)
      rect_brush  = QtGui.QBrush(  QtGui.QColor(self.colors['bkg']) )
      #rect_pen =   QtGui.QPen(    QtGui.QColor(self.colors['bkg']),  self.pen_size, QtCore.Qt.SolidLine)

      y1_arrow_trunk=  (self.height - self.arrow_height) / 2.0
      y2_arrow_trunk=  y1_arrow_trunk + self.arrow_height
      y1_arrow_head =  0
      y2_arrow_head =  self.height
      y_arrow_head=    self.height/2.0
      arrow_trunk_width= self.width - self.pointer_width
      x1_arrow_head=  arrow_trunk_width
      x2_arrow_head=  self.width
      x_arrow_trunk = 0
        
      if (self.gene.strand=='-' and not self.reverse)    or   (self.gene.strand=='+' and self.reverse) :
        y1_arrow_trunk, y2_arrow_trunk= y2_arrow_trunk, y1_arrow_trunk
        y1_arrow_head, y2_arrow_head  = y2_arrow_head, y1_arrow_head
        x_arrow_trunk = self.width
        x1_arrow_head = self.pointer_width
        x2_arrow_head = 0
      
      qpoints_list= [    #adding the seven points to contruct the arrow. starting from top left corner (when plus strand)
                 QtCore.QPointF(   x_arrow_trunk,  y1_arrow_trunk)  , #A
                 QtCore.QPointF(   x_arrow_trunk,  y2_arrow_trunk)  , #B
                 QtCore.QPointF(   x1_arrow_head,  y2_arrow_trunk)  , #C
                 QtCore.QPointF(   x1_arrow_head,  y2_arrow_head)   , #D
                 QtCore.QPointF(   x2_arrow_head,  y_arrow_head)    , #E -- pointer
                 QtCore.QPointF(   x1_arrow_head,  y1_arrow_head)   , #F
                 QtCore.QPointF(   x1_arrow_head,  y1_arrow_trunk)  ] #G

      if self.gene.strand is None:   ## just a rectangle
        qpoints_list = [ QtCore.QPointF(x_arrow_trunk,  y1_arrow_trunk), QtCore.QPointF(x_arrow_trunk,  y2_arrow_trunk),  
                         QtCore.QPointF(x2_arrow_head,  y2_arrow_trunk), QtCore.QPointF(x2_arrow_head,  y1_arrow_trunk)   ]

      arrow=   QtGui.QGraphicsPolygonItem(   QtGui.QPolygonF(  qpoints_list ),   self.item)  
      arrow.setPen(outline_pen)
      arrow.setBrush(rect_brush)      
      self.draw_text_fields()
      return  
 
def syntheny_view ( list_of_genes, get_color= lambda x:x.color, face=ArrowGeneFace, margins={'margin_left':4, 'margin_right':4, 'margin_top':4, 'margin_bottom':4 }, reverse=False,  **other_face_attributes):
  """  Function that accepts a list of genes in input (MMlib.gene class) indicating a synthenic block  and returns a list of ArrowGeneFace faces to represent this.
To represent things other than genes, you can include any of these special values as elements in the input list of genes:
   '-'      :  gaps in the syntheny  (-> GapFace)
   ''       :  white spacer          (-> GapFace  with color=None)
   'start'  :  contig starts here    (-> ContigTerminusFace with mode='start')
   'start'  :  contig starts here    (-> ContigTerminusFace with mode='start')
This method should be called for each node in which you want to have this view. The representation of genes in different node can be conceptually linked through colors.
The function get_color is a generalized way to derive the color of each gene. This function should return a color name. The easiest way to define colors is to add attributes to each gene object. The attribute .color will define the background of the arrow. Then, four special attributes in the gene objects are recognized: .color_outline, .color_text, .color_box_bkg, .color_box_line
The argument margins is a hash defining the margin attributes set in the faces created.
If reverse==True, then the representation is flipped. This is typically to represent the context of a gene which is on the negative strand in the assembly.
  other_face_attributes:  other keywords can be specified setting any attribute of the face (e.g. see ArrowGeneFace.default_attributes). All attributes specified like this are set for all genes in list_of_genes.
  NOTE: it is also possible to specify attributes in the input gene objects, allowing to customize them one by one. The attribute "face_attributes" must be defined in the gene object; this is a hash in the same format as you would put other_face_attributes. 
   E.g. g=gene(strand='+', chromosome='chr1', face_attributes ={'font':'Arial', 'font_size':5}  )    --> then provide g in list_of_genes
       
Usage example: 
for col_index, the_face in enumerate( syntheny_view( [ gene(strand='+', id='gene1',   color='blue', color_box_bkg='yellow'), gene(strand='-', id='gene2', color='green',     color_outline='black'  )])):   node.add_face( the_face, col_index+1, 'aligned')

 """
  #face_container= FaceContainer()
  face_list=[]
  for g_index, g in enumerate(list_of_genes):
    if 'width' in  other_face_attributes: width=other_face_attributes['width']
    else:                                 width=face.default_attributes['width'] 
    if 'height' in  other_face_attributes: height=other_face_attributes['height']
    else:                                  height=face.default_attributes['height'] 

    if      g is None or g=='-':  agf=GapFace(width=width, height=height)
    elif    g == '':              agf=GapFace(width=width, height=height, color=None)
    elif g in ['start', 'end']:   agf=ContigTerminusFace(width=width, height=height, mode=g) 
    else:
      try:     
        bkg_color= get_color( g )
        assert bkg_color
      except:  
        printerr("ERROR syntheny_view the function get_color provided failed with this gene: "+str(g.id), 1)
        raise

      if hasattr(g, 'color_outline') and g.color_outline: outline_color=g.color_outline
      else:                                               outline_color=bkg_color
      if hasattr(g, 'color_text') and g.color_text: text_color=g.color_text
      else:                                         text_color='black'
      if hasattr(g, 'color_box_bkg')  and g.color_box_bkg:   box_bkg_color=g.color_box_bkg
      else:                                                  box_bkg_color= None
      if hasattr(g, 'color_box_line') and g.color_box_line:  box_line_color=g.color_box_line
      else:                                                  box_line_color=None

      colors={'bkg':bkg_color, 'outline':outline_color, 'text':text_color, 'box_bkg':box_bkg_color, 'box_line':box_line_color  }      
      #print colors
      
      other_face_attributes_for_this= dict(other_face_attributes) #copying dict
      if hasattr(g, 'face_attributes'): 
        for k in g.face_attributes:          
          if    k=='width':  width= g.face_attributes[k]
          elif  k=='height': height=g.face_attributes[k]
          elif  k=='colors': colors=g.face_attributes[k]
          else: other_face_attributes_for_this[k] = g.face_attributes[k]
      for k in ['width', 'height', 'colors']: 
        if k in other_face_attributes_for_this: del other_face_attributes_for_this[k]

      agf=face(g, colors=colors, width=width, height=height, reverse=reverse, **other_face_attributes_for_this)
    for k_margin in margins:      setattr(agf, k_margin, margins[k_margin])
    if reverse:   face_list.insert(0, agf)
    else:         face_list.append(agf)
  return face_list


class ExonViewFace(GeneFace):
    """ Exon centric representation of a gene with potentially multiple exons, as a colored rectangle. Introns are compressed to vertical lines, and gaps are permitted so that multiple instances of this could represent a protein alignment. Attributes may be printed within the rectangle.
  -Init: ExonViewFace(   gene_obj, gaps="", printed={}, annotations={}, **attributes   )
    *gene object: must have a strand, a chromosome, and list of exon positions ( .exons attribute of gene, like [ [st, end], [st2, end2], ...  ]  )
    *gaps: three types of arguments can be provided here; 
      - a list like [ (pos_of_gap_in_nucleotide, gap_length) ]
      - a string that, removing gaps (-), has the same length as the gene object (~ aligned nucleotides); gaps are indicated with "-", the other chars are ignored
      - similarly, a string with sequence of length like gene_length/3, (~aligned proteins)
    *printed:   see GeneFace help
    *annotations:   a hash describing the information of any positional feature that you want to be displayed. format: { 'annotation_name':  {'map': 'MAP_TYPE',  'start':[3, 40, 80], 'end':[8, 46, 81]} } ; additionally: it can also have y1, and it can also have y0.   y0-y1 define the vertical space that is span by this feature. In practice it can be used to display another layer of information for each feature, like a score. Note: y1 and y0 are floats between 0 and 1. 0 means the inferior baseline, 1 the superior baseline.
      - MAP_TYPE defines the relation between the positions given and the actual drawing position; note that positions are always 1-based, nucleotide based; value "SEQ" --> the positions are relative to the single sequence (e.g. 1 is first position of gene object); value "ALI" --> the positions are relative to the alignment, meaning, gaps are also counted as possible positions (e.g. 1 is always the first position in the drawing, the extreme left of the rectangle)
      - 'start' and 'end' provide the list of positions. They should have the same number of elements. In case the annotation refers to a single point, 'end' can be omitted, and in this case it's like end= start for each of the annotation instances.
      (color of the annotations can be defined by providing through the "colors" argument of __init__: just use as key the annotation name. If annotations spans more than a single position ('end' defined) the color defines the pen outline and also the brush fill; otherwise, only pen is defined)
      the annotation names 'introns' is reserved for drawing the introns as vertical lines. The value must be True to display.   ###and 'frameshifts' and stop
    *colors:    see GeneFace help
    *other attributes: see GeneFace help; list: order_text, height, width OR width_per_position, font, font_size, shrink_font_to_fit, font_margin, rotable
"""
    default_attributes={
      'order_text':['id', 'chromosome', 'boundaries', 'text'],
      'height':40, 'width':150, 'width_per_position':None, 'min_for_gaps':5,
      'font':'Courier New', 'font_size':12, 'shrink_font_to_fit':True, 'font_margin':4,
      'colors':{'bkg':'#4444AA', 'outline':'#000000', 'text':'#000000', 'gap':'#999999', 'introns':'#000000', 'frameshifts':'#EE2222' },
      'printed':{'id':1, 'chromosome':1, 'text':0, 'boundaries':1},
      'annotations':{'introns':True, 'frameshifts':True},
      'pen_size':1,
      'rotable':False,
    }

    def __init__(self, gene_obj, gaps="", annotations=None, **attributes):
      GeneFace.__init__(self, gene_obj, **attributes)      ### attributes are added / updated here

      ###  processing gaps position, by joining them
      self.gaps=[]    #form:    [gap_relative_pos, gap_length]   relative pos in in nucleotides, 0 means before the first, 1 after the first etc          
      if type(gaps)==str and len(nogap(gaps))*3  == len(self.gene) : # if variable gaps is provided as aligned proteins
        gaps=''.join([ 3*char   for char in gaps ])  ## modifying gaps to enter in next "if"
      if type(gaps)==str and len(nogap(gaps)) == len(self.gene): # if variable gaps is provided as aligned nucleotides
        index=0; tot_gaps=0 #increasing
        while index<len(gaps):
          if gaps[index]=='-': 
            gap_start=index
            while index<len(gaps) and gaps[index]=='-':               index+=1
            gap_length=index-gap_start
            self.gaps.append([gap_start-tot_gaps, gap_length])
            tot_gaps+=gap_length            
          index+=1
      elif type(gaps)==list: self.gaps=gaps ## this was provided directly in the form it is stored (See above)
      else:         raise Exception("ERROR initializing ExonViewFace (id={0}) the gaps argument was not recognized! wrong length maybe? \ngaps={1}\nlen(gaps)={2} len(gaps without '-')={3}  gene.length()={4}".format(self.gene.id, gaps, len(gaps), len(nogap(gaps)), self.gene.length() ))
      # add gaps of length 0 at beginning and end, if not present; this allows to have a nice data structure for later parsing
      if not self.gaps or self.gaps[0][0] != 0:   self.gaps.insert(0, [0, 0])  
      if self.gaps[-1][0] != self.gene.length():  self.gaps.append([self.gene.length(), 0])        

      ### get or set self.width and self.width_per_position
      if self.width_per_position:     self.width= self.width_per_position * (  self.gene.length() + sum([gap_length for g_p, gap_length in self.gaps])  )
      elif self.width:                self.width_per_position=float(self.width)/ ( self.gene.length() + sum([gap_length for g_p, gap_length in self.gaps]) )
      else: raise Exception("ExonViewFace ERROR neither width or width_per_position are set!")

      ### filtered gaps: precomputed minimal data structure on gaps positiions and length
      self.filtered_gaps=[]  #pos, length, gap_length_before_this   #### I need to keep memory of the total size of all gaps -- including those too short to be displayed
      total_gap_length=0   ## in nucleotides, progressively increased      
      min_length_gap_to_be_displayed= self.min_for_gaps / self.width_per_position
      for pos, length in self.gaps:
        if (length >=min_length_gap_to_be_displayed) or (pos in [0, self.gene.length() ]) :
          self.filtered_gaps.append(  [pos, length, total_gap_length]  )
        total_gap_length+=length

      #### PARSE ANNOTATIONS; make everything nucleotide, alignment based
        # preparing "annotations";  filling default values 
      if annotations is None: annotations={}
      for key in  self.default_attributes['annotations']:  
        if key not in annotations: annotations[key]= self.default_attributes['annotations'][key]
      #  print "pre: ", annotations, '\n', self.annotations
      ## pre processing special annotation: introns
      if ('introns' in annotations and annotations['introns']) or ('frameshifts' in annotations and annotations['frameshifts']):
        introns_dict=    {'map':'SEQ', 'start':[], 'end':[]}
        frameshifts_dict={'map':'SEQ', 'start':[], 'end':[]}
        if len(self.gene.exons)>=2:  
          nucleotides_behind=0
          for exon_index in range( 1, len(self.gene.exons) ):
            previous_exon_length= self.gene.exons[exon_index-1][1] - self.gene.exons[exon_index-1][0]
            if    self.gene.strand=='+':    intron_length= self.gene.exons[exon_index][0] - self.gene.exons[exon_index-1][1] - 1 
            elif  self.gene.strand=='-':    intron_length= self.gene.exons[exon_index-1][0] - self.gene.exons[exon_index][1] - 1 
            nucleotides_behind+=previous_exon_length
            if intron_length < 6 : 
              frameshifts_dict['start'].append( nucleotides_behind )
              frameshifts_dict['end'].append( nucleotides_behind )
            else: 
              introns_dict['start'].append( nucleotides_behind )
              introns_dict['end'].append( nucleotides_behind )
        if 'introns' in annotations and annotations['introns']  and introns_dict['start']:   #it was requested, and there's at least one intron to display
          annotations['introns']= introns_dict
        elif   'introns' in annotations:           del annotations['introns']          

        if 'frameshifts' in annotations and annotations['frameshifts']  and frameshifts_dict['start']:   #it was requested, and there's at least one fs to display
          annotations['frameshifts']= frameshifts_dict
        elif 'frameshifts' in annotations:           del annotations['frameshifts']          
                       
      ## general parse annotations; putting everything in alignment relative coordinates; generating and filling self.annotations
      self.annotations={}
      for annotation_name in annotations:
        if not annotations[annotation_name]: 
          self.annotations[annotation_name]=False
          continue
        ann_dict = annotations[annotation_name]
        if 'end' in ann_dict and len(ann_dict['start'])!= len(ann_dict['end']): raise Exception("ExonViewFace ERROR annotation name: "+annotation_name+" start and end index lists have different length! start: "+str(ann_dict['start'])+' end: '+str(ann_dict['end']))
        elif 'end' not in ann_dict:  ann_dict['end']=ann_dict['start']
        ### checking y1 and y0
        if 'y1' not in ann_dict: ann_dict['y1']=  [1.0 for index in range( len(ann_dict['start']) ) ]
        elif len(ann_dict['start'])!= len(ann_dict['y1']): raise Exception("ExonViewFace ERROR annotation name: "+annotation_name+" start and y1 values lists have different length! start: "+str(ann_dict['start'])+' y1: '+str(ann_dict['y1']))
        if 'y0' not in ann_dict: ann_dict['y0']=  [0.0 for index in range( len(ann_dict['start']) ) ]
        elif len(ann_dict['start'])!= len(ann_dict['y0']): raise Exception("ExonViewFace ERROR annotation name: "+annotation_name+" start and y0 values lists have different length! start: "+str(ann_dict['start'])+' y0: '+str(ann_dict['y0']))          

        if any( [  y0_value<0.0 or y0_value>1.0  for y0_value in ann_dict['y0'] ] ): 
          raise Exception("ExonViewFace ERROR annotation name: "+annotation_name+" y0 values not valid! they must be between 0.0 and 1.0! "+str(ann_dict['y0']))
        if any( [  y1_value<0.0 or y1_value>1.0  for y1_value in ann_dict['y1'] ] ):
          raise Exception("ExonViewFace ERROR annotation name: "+annotation_name+" y1 values not valid! they must be between 0.0 and 1.0! "+str(ann_dict['y1']))
        
        if ann_dict['map']=='ALI':            self.annotations[annotation_name]={'start':ann_dict['start'], 'end':ann_dict['end'],  'y0':ann_dict['y0'], 'y1':ann_dict['y1']  }
        elif ann_dict['map']=='SEQ':          self.annotations[annotation_name]={'start': list(map(self.position_in_ali, ann_dict['start'])), 'end':list(map(self.position_in_ali, ann_dict['end'])), 'y0':ann_dict['y0'], 'y1':ann_dict['y1'] }
        else: raise Exception("ExonViewFace ERROR map argument not recognized: '"+ann_dict['map']+"'")
        ## checking annotation positions  ## plus correcting for gap positions: splitting in di
        ali_tot_length=self.ali_length()
        index=0  #annotation idnex
        gap_index=0  #to avoid reparsing almost from first gap
        while index < len(self.annotations[annotation_name]['start']):
          start=self.annotations[annotation_name]['start'][index]          
          end=  self.annotations[annotation_name]['end'][index]          
          if max( (start, end) ) > ali_tot_length:
            #raise Exception, 
            printerr("WARNING ExonViewFace ERROR annotation_name: "+annotation_name+" annotation_index: "+str(index)+ " Feature out of boundaries! Total positions: "+str(ali_tot_length)+'  Start, end for this annotation were: '+str((start, end))    , 1)         
          while gap_index < len(self.filtered_gaps):
            gap_start_in_seq, gap_length, gaps_behind = self.filtered_gaps[gap_index]
            pos_in_ali=gap_start_in_seq + gaps_behind  #start position of gap in alignment coordinates
            if pos_in_ali<start: gap_index+=1; continue
            if pos_in_ali>end: break
            #    there's a gap in the middle of the domains coordinate. Let's split the domain in pieces so that it is drawn correctly
            self.annotations[annotation_name]['end'].insert(index, pos_in_ali)
            self.annotations[annotation_name]['start'].insert(index+1, pos_in_ali+gap_length)
            self.annotations[annotation_name]['y0'].insert(index+1,  self.annotations[annotation_name]['y0'][index]    )
            self.annotations[annotation_name]['y1'].insert(index+1,  self.annotations[annotation_name]['y1'][index]    )
            gap_index+=1;     
            end=pos_in_ali            
          index+=1
            
            #if start > ali_tot_length: start =ali_tot_length   ### ???
            #if end > ali_tot_length:   end   =ali_tot_length   ###
      #print   "# !! ", self.gene.id, "self.annotations", self.annotations
          
    def position_in_ali(self, pos_in_seq):
      """ Analogous to the function in class alignment; parses self.gaps to translate a sequence relative position to a alignment postion. Evrything is one based"""
      if type(pos_in_seq)!=int: raise Exception("ERROR id "+self.gene.id+' position_in_ali called with something else than a integer: '+str(pos_in_seq)+' ('+str(type(pos_in_seq))+')')
      gaps_behind=0; gap_index=0; 
      #print [pos_in_seq], self.gaps
      while  self.gaps and  gap_index<len(self.gaps) and self.gaps[gap_index][0]  < pos_in_seq:
          #print "ok: ", pos_in_seq, gap_index, self.gaps[gap_index], self.gaps[gap_index][0], " < ", pos_in_seq, self.gaps[gap_index][0]  < pos_in_seq
        gaps_behind+=self.gaps[gap_index][1]
        gap_index+=1
      return pos_in_seq+gaps_behind

    def ali_length(self):
      return self.gene.length() + sum(  [g_length   for g_p, g_length in self.gaps]  )
        
    def update_items(self):
      """ the actual function doing the graphics """
      self.item = QtWidgets.QGraphicsRectItem(0, 0, self.width, self.height) # backbone object, no color 
      self.item.setPen(QtGui.QPen(QtCore.Qt.NoPen)) #no black border      

      gap_pen     = QtGui.QPen(    QtGui.QColor(self.colors['gap']),      self.pen_size, QtCore.Qt.DotLine)
      outline_pen = QtGui.QPen(    QtGui.QColor(self.colors['outline']),  self.pen_size, QtCore.Qt.SolidLine)
      rect_brush  = QtGui.QBrush(  QtGui.QColor(self.colors['bkg']) )
      rect_pen    = QtGui.QPen(    QtGui.QColor(self.colors['bkg']),      self.pen_size, QtCore.Qt.SolidLine)           

      for gap_index, gap_item in enumerate(self.filtered_gaps): #between a gap and the next one   ### artificial gaps are present here at pos 0 and pos length.
        gap_relative_pos, gap_length, total_gap_length= gap_item  ## relative pos is the pos of the gap in the target, nucleotide based; 0 means before the first nucleotide. 1 after the first nucleotide, etc             
        ### draw gap
        if gap_length: ## in filtered_gaps, we have either gaps of an accepted (>threshold) length, or gaps of length 0, as first or last element, just for structure
          x_start=(total_gap_length + gap_relative_pos)*self.width_per_position
          x_end=   x_start + self.width_per_position*gap_length -1        # -1 to avoid overlaps                     
          if x_end > x_start:   # basically all the times. But if we have a very small gap_length at the beginning, this could give a negative x_end, so we have this control
            line_above=QtWidgets.QGraphicsLineItem( x_start, 0,           x_end, 0,           self.item) 
#          line_below=QtWidgets.QGraphicsLineItem( x_start, self.height, x_end, self.height, self.item) 
            line_above.setPen(gap_pen);          #line_below.setPen(gap_pen)

        ### drawing the colored rectangle. We do this for every gap except the last one
        if gap_index!= len(self.filtered_gaps)-1:    
          region_length  = self.filtered_gaps[gap_index+1][0] - gap_relative_pos #nucleotides
          region_length += self.filtered_gaps[gap_index+1][2] - self.filtered_gaps[gap_index][2] - self.filtered_gaps[gap_index][1]      #adding the small insertions in this cds; this is equal to the differences of the  total number of gaps recorded in the next gap and in this one, minus the length of this gap.          
          x_start=(total_gap_length + gap_relative_pos + gap_length  )*self.width_per_position
          width  = self.width_per_position*region_length -1    # -1 to avoid overlap 
          x_end  = x_start + width
          # filled rectangle first, then one line above and one below
          colored_rect = QtWidgets.QGraphicsRectItem(x_start, 0, width, self.height, self.item)   
          colored_rect.setPen(rect_pen)
          colored_rect.setBrush(rect_brush)
          line_above=QtWidgets.QGraphicsLineItem( x_start, 0,           x_end, 0,           self.item) #parent=
          line_below=QtWidgets.QGraphicsLineItem( x_start, self.height, x_end, self.height, self.item) #parent=
          line_above.setPen(outline_pen);          line_below.setPen(outline_pen)
      self.draw_text_fields()
      self.draw_annotations()
      return  

    def draw_annotations(self):
      sorted_annotation_names=sorted([a  for a in self.annotations if a != 'introns' ])  # making sure introns are drawn last
      if 'introns' in self.annotations: sorted_annotation_names.append('introns')
      for annotation_name in sorted_annotation_names:
        #print "ANNOTATION   ", annotation_name, self.annotations[annotation_name]
        if self.annotations[annotation_name]:     ## to be able to turn off default tracks we add this control
          if annotation_name in self.colors:  color=self.colors[annotation_name]
          else:                                     color='#000000'
          pen_this_annotation=    QtGui.QPen(     QtGui.QColor(color) )
          brush_this_annotation=  QtGui.QBrush(   QtGui.QColor(color) )
          for ann_index, ann_start in enumerate(self.annotations[annotation_name]['start']):
            ann_end= self.annotations[annotation_name]['end'][ann_index]
            x_start= (ann_start)*self.width_per_position
            x_end=   (ann_end)  *self.width_per_position
            try:
              y0_relative=self.annotations[annotation_name]['y0'][ann_index]
              y1_relative=self.annotations[annotation_name]['y1'][ann_index]
            except IndexError:
              #printerr(  (annotation_name, ann_index, ann_start, self.annotations[annotation_name]['y0']), 1)
              write(   (annotation_name, ann_index, ann_start, self.annotations[annotation_name]['y0']), 1)
              y0_relative=0
              y1_relative=1
              
            y0_absolute=self.height*(1.0 - y0_relative)      #turning upside down as qt4 likes
            y1_absolute=self.height*(1.0 - y1_relative)      #turning upside down as qt4 likes
            
            y_min  =y1_absolute;    y_max=y0_absolute

            if x_start != x_end:
              ## drawing rectangle
              colored_rect = QtWidgets.QGraphicsRectItem(x_start, y_min, x_end-x_start    , y_max-y_min, self.item)   ## no border
              colored_rect.setPen(pen_this_annotation)
              colored_rect.setBrush(brush_this_annotation)
            else: 
              ## drawing line
              colored_line= QtWidgets.QGraphicsLineItem( x_start, y_min, x_end, y_max, self.item)
              colored_line.setPen(pen_this_annotation)


class SmallExonViewFace(ExonViewFace):
    default_attributes={
      'order_text':['id', 'chromosome', 'boundaries', 'text'],
      'height':20, 'width':100, 'width_per_position':None, 'min_for_gaps':6,
      'font':'Courier New', 'font_size':12, 'shrink_font_to_fit':True, 'font_margin':4,
      'colors':{'bkg':'#4444AA', 'outline':'#000000', 'text':'#000000', 'gap':'#999999', 'introns':'#000000', 'frameshifts':'#EE2222' },
      'printed':{'id':0, 'chromosome':1, 'text':0, 'boundaries':0},
      'annotations':{'introns':True, 'frameshifts':True}
    }


#print globals()

"""
def remember():                                                
#gene_brick_height/6, offset_for_id, gene_brick_height*2/3) #parent item: bkg

      offset_for_additional_here=offset_for_additional*len(self.gene.additional_features)      


      if not opt['no_id']:
        ### rectangle for prediction id
        chrom_id_rect = QtWidgets.QGraphicsRectItem(0, gene_brick_height/6, offset_for_id, gene_brick_height*2/3) #parent item: bkg
        chrom_id_rect.setParentItem(self.item)
        ## text for prediction id
        obj=QtGui.QGraphicsSimpleTextItem() 
        obj.setText(self.gene.id)
        shrink_font_to_box(obj, offset_for_id, gene_brick_height*2/3, category='prediction id')  ## modifying font size in place
        

        obj.setPos( (offset_for_id-obj.boundingRect().width())/2 , (gene_brick_height-obj.boundingRect().height())/2 ) #centering the text
        obj.setParentItem(chrom_id_rect)

      for index, x in enumerate( self.gene.additional_features ):
        #write(self.gene.full_id+' drawing secis '+' '+str(offset_for_additional), 1)
        secis_rect = QtWidgets.QGraphicsRectItem(gene_brick_width+offset_for_id+offset_for_additional*index, gene_brick_height/8, offset_for_additional-1, gene_brick_height*6/8) #parent item: bkg
        secis_rect.setParentItem(self.item)
        secis_rect.setBrush(QtGui.QBrush(QtGui.QColor( x.color()    )))
        #pen=QtGui.QPen(); pen.width=3
        #pen.setColor(QtGui.QColor("#FFFFFF"))
        #secis_rect.setPen(QtGui.QPen(color="#FFFFFF", width=2) )
        obj=QtGui.QGraphicsSimpleTextItem()         
        obj.setText(x.text)
        obj.setBrush( QtGui.QBrush(QtGui.QColor( "#FFFFFF"  ) ))
        reduce_font_if_necessary(obj, offset_for_additional_here, gene_brick_height*6/8, category='secis description')
        obj.setParentItem(secis_rect)
        obj.setPos(  gene_brick_width+offset_for_id+offset_for_additional*index+ (offset_for_additional-obj.boundingRect().width())/2 , (gene_brick_height-obj.boundingRect().height())/2 ) #centering the text

      #drawing line to represent the 100 coverage for profile
      line_for_full_coverage= QtWidgets.QGraphicsLineItem(offset_for_id, 0, offset_for_id+gene_brick_width, 0)
      line_for_full_coverage.setParentItem(self.item)
      ## rectangle for gene brick
      gene_brick=QtWidgets.QGraphicsRectItem(offset_for_id+self.gene.relative_boundaries()[0]*gene_brick_width, 0, (self.gene.relative_boundaries()[1]-self.gene.relative_boundaries()[0])*gene_brick_width, gene_brick_height) 
      gene_brick.setBrush(QtGui.QBrush(QtGui.QColor(self.gene.color())))
      gene_brick.setParentItem(self.item)

      #drawing lines for introns
      if not opt['I'] and len(self.gene.exons)>1:
        cds_so_far=0
        tot_cds=self.gene.length()
        for exon_index in range(len(self.gene.exons[:-1])):
          start, end = self.gene.exons[exon_index]
          cds_so_far+= end-start+1
          aa_so_far = cds_so_far/3.0
          if self.gene.strand=='+':     intron_length= self.gene.exons[exon_index+1][0] - end -1   #######
          elif self.gene.strand=='-':   intron_length= start - self.gene.exons[exon_index+1][1] -1   #######
          x_intron  =  self.gene.relative_position_in_ali_of(aa_so_far) * gene_brick_width  +  offset_for_id
          line_intron= QtWidgets.QGraphicsLineItem( x_intron, 1, x_intron, gene_brick_height-2)
          
          color=opt['intron_color'] # '#FFFFFF' #white
          if intron_length<=5: color='#EE0000' #red for frameshifts
          line_intron.setPen(QtGui.QPen(QtGui.QColor(color)))
          line_intron.setZValue(1)
          line_intron.setParentItem(gene_brick)          

      if opt['f']:
        for feature_name in self.gene.graphical_features: 
          for aa_position in self.gene.graphical_features[feature_name]:
            x_feature= self.gene.relative_position_in_ali_of(aa_position) * gene_brick_width  +  offset_for_id
            line_feature = QtWidgets.QGraphicsLineItem( x_feature, 1, x_feature, gene_brick_height-2)
            if graphical_features_colors.has_key(feature_name):
              line_feature.setPen(QtGui.QPen(QtGui.QColor( graphical_features_colors[feature_name]  )))
            line_feature.setZValue(1.1)
            line_feature.setParentItem(gene_brick)
            
      if not opt['T']:
      ## text for chromosome
        obj=QtGui.QGraphicsSimpleTextItem() 
        obj.setPos(offset_for_id+1, 0)
        obj.setText(self.gene.chromosome)
        font=obj.font();  font.setPointSize(opt['bsize']);     obj.setFont(font) #setting default text size
        reduce_font_if_necessary(obj, gene_brick_width, category='text for chromosome')
        obj.setZValue(2)
        obj.setParentItem(gene_brick)
      ## text for positions
        obj=QtGui.QGraphicsSimpleTextItem() 
        obj.setPos(offset_for_id+1,gene_brick_height/2 +1 )
        obj.setText(join([str(i) for i in self.gene.boundaries()], self.gene.strand)) # e.g. 1+101 (positive strand) 45-80 (negative strand
        font=obj.font();  font.setPointSize(opt['bsize']);     obj.setFont(font) #setting default text size      
        reduce_font_if_necessary(obj, gene_brick_width, category='positions text')
        obj.setZValue(2.1)
        obj.setParentItem(gene_brick)

    def _width(self):        return self.item.rect().width()
    def _height(self):       return self.item.rect().height()

"""



#! /usr/bin/env python3
#from string import *
import sys, os, shlex, subprocess
from easyterm import *
from Bio import SeqIO
from glob import glob
from . import alignment_tools


##### catching up from MMlib
def parse_fasta(filename):
  with open(filename) as fh:
    for title, seq in SeqIO.FastaIO.SimpleFastaParser(fh):
      yield (title, seq)

def fasta(seq):
  return '\n'.join(  [chunk for chunk in seq[::60] ] )

def bbash(cmnd, print_it=0,  out=None, err=None, tolerate=False, shell=False):
  if out is None and err is None:
    out=subprocess.PIPE
    err=subprocess.STDOUT
  elif out is None or err is None:
    raise Exception('bbash ERROR if you specify out then specify err too')
  p=subprocess.run( shlex.split(cmnd),
                  stdout=out, stderr=err,
                    check=False, shell=shell)
  o=p.stdout  
  if not tolerate and p.returncode!=0:
    if o:
      raise Exception("bbash ERROR command: {} error: {}".format(cmnd, o.decode(encoding = 'utf-8')))
    else: 
      raise Exception("bbash ERROR command: {} error --> check {}".format(cmnd, o.__file__))
  if o:
    return o.decode(encoding = 'utf-8')
  else:
    return o

help_msg=""" Program to build and draw a tree for a set of proteins together with their most similar annotated proteins identified by blast against a target DB.

Usage:  phylo_context  -i protein_sequences.fasta  -o output_prefix  -db protein_db.fa

* Compulsory I/O:

-i    path to input protein sequences in fasta format
-db   path to blast protein database, must be already formatted with makeblastdb
-o    this argument will be used as prefix for all output files

* You must also include one of these options:
-st    file defining species name for each sequence id; such as 'Homo sapiens'  or 'Homo_sapiens'
        format:  seqid -tab- species
-sf    python code to derive species from sequence id; example: x.split('.')[3]
Note that the species label 'ncbi' can be used to fetch species name online; it works for NCBI ids

* All these routines are executed sequentially:
B = Blastp+ of all proteins against a reference database (ncbi NR recommended)
 -mts  Maximum number of aligned sequences to keep  [6000]
 -rem  Run blastp remotely; please use -db nr (or another accepted db name) with this

P = Parse blastp output and fetch proteins from reference database
 -e    evalue filter of blast hits              [1e-5]
   
R = align "context" proteins (those matched in ref db) and Reduce them to a feasible number
 -m    maximum amount of context proteins       [300]

T = run ete3 build Tree phylogeny on the ali of input and context proteins
 -w    ete3 workflow                            [phylomedb4]
 -c    run with --clear (instead of --resume)

F = build tree figures, one with context protein descriptions and one without
 -nf   do not build figures (just show the command)

Force recomputing any given routine by using the corresponding option: 
 e.g. -B to recompute blastp (and everything after that).

This options allows to reutilize useful output from a previous run:
-load  prev_out_prefix
When using -load, you HAVE to provide the routine option corresponding to where to restart from 
 e.g. -R to not recompute blast, but do recompute context proteins, tree and figures

### other options:
-dr            do not run anything, but print commands    **** not supported !
-n              number of threads used for blastp and mafft   [20]
-print_opt      print currently active options
-h OR --help    print this help and exit"""

command_line_synonyms={}

def_opt= {'temp':'/tmp', 
          'i':'', 'o':'',
          'B':False, 'P':False, 'R':False, 'T':False, 'F':False,
          'db':'undefined_DB',   'e':1e-5,  'n':20, 
          'm':300,     'mts':6000,
          'rem':False,
          'w':'standard_trimmed_fasttree',    
          'dr':False,   'c':False, 'nf':False,
          'load':False,
          'st':'', 'sf':'',
}



#########################################################
###### start main program function
how_routine =''
how_found   ='' #green'
how_main    ='blue'
how_commands='magenta'
how_info=    'yellow'


def main(args={}):  
#########################################################
############ loading options
  if not args:
    opt=command_line_options(def_opt, help_msg, '', synonyms=command_line_synonyms )
  else:
    opt=args
  temp_folder=random_folder(opt['temp'])

  ############################################################
  #### to do:
  #     -check n sequences, warning or stop
  #     -options of ete  clearall or resume
  #     -show tree (and previous = style etc)

  def run(cmnd, *args, **kargs):
    write(cmnd, how=how_commands)
    if not opt['dr']:
      return bbash(cmnd, *args, **kargs)

  input_file=opt['i'];   check_file_presence(input_file, 'input_file')
  db_file=opt['db'];     #check_file_presence(db_file, 'db_file')
  #output_folder=Folder(opt['o'])
  evalue_threshold=opt['e']
  n_threads=opt['n']
  max_seqs=opt['m']
  ete3_workflow=opt['w']
  max_target_seqs=opt['mts']
  load_prefix=('/auisbdUnMatchathackjacble' if not opt['load'] else opt['load']  ).rstrip('.')+'.'

  if opt['o']:    output_prefix=opt['o']
  else:
    x='.'.join(os.path.basename(input_file).split('.')[:-1]) if os.path.basename(input_file).count('.') else os.path.basename(input_file)
    output_prefix='{}/{}.'.format( os.path.dirname(input_file) or  '.', x)
  if not output_prefix.endswith('.'): output_prefix+='.'

  #log_fileh=open('{}phylo_context.log'.format(output_prefix), 'w')
  log_file='{}phylo_context.log'.format(output_prefix)
  set_logfile(log_file)
  write('# ++++++++++++++++++ phylo_context  ++++++++++++++++++ #',  how=how_main)
  write('# start date: {}'.format(bbash('date'))) #, how=how_main)
  write(opt)

  ## checking for duplicates; also, keeping titles (first word) in stitles
  stitles={}
  some_dups=False
  txt_no_dups=''
  for t,seq in parse_fasta(input_file):
    splt=t.split()
    st=splt[0]
    if st in stitles:
      some_dups=True
      i=2
      nt='{}{}'.format(st, i)
      while nt in stitles:
        i+=1
        nt='{}{}'.format(st, i)
      stitles[nt]=st
      st=nt
    else:
      stitles[st]=None
    txt_no_dups+='>{}\n{}\n'.format( ' '.join( [st]+splt[1:] ),   fasta(seq) )
  if some_dups:
    no_dups_file='{}no_dups.fa'.format(output_prefix)
    dups_tsv='{}dups.tsv'.format(output_prefix)
    write('WARNING duplicate titles were detected! removing them. -->  {}    old/new titles  -->  {}'.format(no_dups_file, dups_tsv))
    with open(no_dups_file, 'w') as fh:
      fh.write(txt_no_dups)

    with open(dups_tsv, 'w') as fh:
      fh.write( '\n'.join(['{}\t{}'.format(st, stitles[st])   for st in stitles if not stitles[st] is None]) )

    input_file=no_dups_file    


  if opt['st']:
    seqid2species={s:None  for s in stitles}    
    for line in open(opt['st']):
      seqid, species=line.strip().split('\t')
      if not seqid in seqid2species:
        raise Exception(f"-st file ERROR id not found in sequence file: {seqid}")
      seqid2species[seqid]=species
    not_found=[seqid for seqid in seqid2species if seqid2species[seqid] is None]
    if len(not_found):
      raise Exception(f"-st file ERROR ids not covered:  {' '.join(not_found)}")    
  elif opt['sf']:
    species_function=eval('lambda x:'+opt['sf'])
    seqid2species={s:species_function(s)    for s in stitles}
  else:
    raise Exception("ERROR you must specify option -st or -sf")


  ## routine options
  sequential_routines='BPRTF';  forcing=False
  for r in sequential_routines:
    if opt[r]:    forcing=True
    elif forcing: opt[r]= True

  n_input_seqs=int(bbash('grep -c ">" {}'.format(input_file)))
  write('# input sequences: {}'.format(n_input_seqs), how=how_info)  

  ### add controls! n of seqs vs n of context

  write('#### routine: BLAST        ######', how=how_routine)
  blast_output='{}blastp'.format(output_prefix)
  temp_blast_output='{}blastp.temp'.format(output_prefix)
  load_blast_output='{}blastp'.format(load_prefix)
  if not os.path.isfile(blast_output) or opt['B']:
    if os.path.isfile(load_blast_output) and not opt['B']:
      load_rel_path=smartlink(load_blast_output, blast_output)
      write('# <- link: {} to {}'.format(load_rel_path, blast_output), how=how_found)
      os.symlink(load_rel_path, blast_output)
    else:
      check_file_presence(db_file, 'db_file')
      blast_cmd='blastp -max_target_seqs {maxs}  -db {d} -query {q} -out {to} -outfmt 6 {remote} -num_threads {nt}'.format(
        d=db_file, q=input_file, to=temp_blast_output, nt=n_threads, maxs=max_target_seqs,
        remote='-remote' if opt['rem'] else '')
      run(blast_cmd)
      run('mv {to} {o}'.format(to=temp_blast_output, o=blast_output) )
  else:
    write('# <- found: {}'.format(blast_output), how=how_found)


  write('#### routine: PARSE        ######', how=how_routine)
  blast_ids='{}hits.ids'.format(output_prefix)
  temp_blast_ids='{}hits.ids.temp'.format(output_prefix)
  blast_fasta='{}hits.fa'.format(output_prefix)
  temp_blast_fasta='{}hits.fa.temp'.format(output_prefix)
  load_blast_ids='{}hits.ids'.format(load_prefix)
  load_blast_fasta='{}hits.fa'.format(load_prefix)
  if not os.path.isfile(blast_fasta) or opt['P']:
    if os.path.isfile(load_blast_fasta) and not opt['P']:
      load_rel_path=smartlink(load_blast_fasta, blast_fasta)
      write('# <- link: {} to {}'.format(load_rel_path, blast_fasta), how=how_found)
      os.symlink(load_rel_path, blast_fasta)
      load_rel_path=smartlink(load_blast_ids, blast_ids)
      write('# <- link: {} to {}'.format(load_rel_path, blast_ids), how=how_found)
      os.symlink(load_rel_path, blast_ids)
    else:
      check_file_presence(db_file, 'db_file')
      parse_cmd='''gawk '$(NF-1)<{e} {{ print $2 }}' {i} '''.format(i=blast_output,  e=evalue_threshold )
      with open(temp_blast_ids, 'w') as fh:
          fh.write(   '\n'.join( [i  for i in    set( run(parse_cmd).split('\n') )    if i] )   +'\n')

      run('mv {to} {o}'.format(to=temp_blast_ids, o=blast_ids))
          
      fetch_cmd='blastdbcmd -db {d} -entry_batch {i} -out {to}'.format(i=blast_ids, d=db_file, to=temp_blast_fasta)
      run(fetch_cmd)
      run('mv {to} {o}'.format(to=temp_blast_fasta, o=blast_fasta))
  else:
    write('# <- found: {}'.format(blast_ids), how=how_found)
    write('# <- found: {}'.format(blast_fasta), how=how_found)
    
  n_context_seqs=int(bbash('wc -l {}'.format(blast_ids)).split()[0])
  write('# blast hits: {}'.format(n_context_seqs), how=how_info)  

  write('#### routine: REDUCE       ######', how=how_routine)
  context_sequences='{}context.fa'.format(output_prefix)
  ali_w_context='{}w_context.fa'.format(output_prefix)
  if n_context_seqs>max_seqs:
    mafft_out='{}hits.mft.fa'.format(output_prefix)
    mafft_temp='{}hits.mft.temp'.format(output_prefix)
    reduce_out='{}hits.mft.r{}.fa'.format(output_prefix, max_seqs)
    reduce_temp='{}hits.mft.r.temp.fa'.format(output_prefix)
    load_mafft_out='{}hits.mft.fa'.format(load_prefix)
    load_reduce_out='{}hits.mft.r{}.fa'.format(load_prefix, max_seqs)
    
    if not os.path.isfile(reduce_out) or opt['R']:
      if os.path.isfile(load_reduce_out) and not opt['R']:
        load_rel_path=smartlink(load_mafft_out, mafft_out)
        write('# <- link: {} to {}'.format(load_rel_path, mafft_out), how=how_found)
        os.symlink(load_rel_path, mafft_out)
        load_rel_path=smartlink(load_reduce_out, reduce_out)
        write('# <- link: {} to {}'.format(load_rel_path, reduce_out), how=how_found)
        os.symlink(load_rel_path, reduce_out)
        
      else:      

        #mafft_cmd='align_with_mafft.py {i} -temp {tmp}  -n {nt} > {to} && mv {to} {o}'.format( i=blast_fasta, nt=n_threads, to=mafft_temp, o=mafft_out, tmp=temp_folder)
        with open(mafft_temp, 'w') as fh:
          mafft_cmd='mafft --anysymbol --auto --thread {nt} {i}'.format( i=blast_fasta, nt=n_threads)        
          run(mafft_cmd, out=fh, err=subprocess.DEVNULL)
        run('mv {to} {o}'.format(to=mafft_temp, o=mafft_out))
        
        #reduce_cmd='alignment_tools -temp {tmp}  -i {i} -repr {m} -o {to}'.format(i=mafft_out, m=max_seqs, to=reduce_temp, tmp=temp_folder)
        #run(reduce_cmd)
        alignment_tools.main({'temp':temp_folder, 'i':mafft_out, 'repr':max_seqs, 'o':reduce_temp})
        
        run('mv {to} {o}'.format(to=reduce_temp, o=reduce_out))

        link_cmd='ln -fs {} {}'.format(os.path.basename(reduce_out), context_sequences)
        run(link_cmd)
        #cat_cmd='nogap.g {} {} > {}'.format(input_file, context_sequences, ali_w_context)
        with open(ali_w_context, 'w') as fh:
          for ffile in [input_file, context_sequences]:
            for t,seq in parse_fasta(ffile):
              fh.write(f">{t}\n" + fasta(seq.replace('-', '')) + "\n")
          #run('nogap.g {} {}'.format(), out=fh, err=fh)

    else:
      write('# <- found: {}'.format(mafft_out), how=how_found)
      write('# <- found: {}'.format(reduce_out), how=how_found)

  else:
    write('# reduce routine not necessary', how=how_info)
    if not os.path.isfile(ali_w_context) or opt['R']:
      link_cmd='ln -fs {} {}'.format(os.path.basename(blast_fasta), context_sequences)
      run(link_cmd)
      with open(ali_w_context, 'w') as fh:
        for ffile in [input_file, context_sequences]:
          for t,seq in parse_fasta(ffile):
            fh.write(f">{t}\n" + fasta(seq.replace('-', '')) + "\n")
        # run('nogap.g {} {}'.format(input_file, context_sequences), out=fh, err=fh)


  ete_option='--resume'  if not opt['c'] else '--clearall'
  
  write('#### routine: TREE        ######', how=how_routine)
  ete_out_folder='{}{}'.format(output_prefix, ete3_workflow)
  if not os.path.isdir(ete_out_folder):
    os.mkdir(ete_out_folder)
  
  ete_tree_link= '{}tree'.format(output_prefix)
  ete_log='{}/ete_build.log'.format(ete_out_folder)
  load_ete_tree_link='{}tree'.format(load_prefix)
  
  if not os.path.isfile(ete_tree_link) or opt['T']:
    if os.path.isfile(load_ete_tree_link) and not opt['T']:
      load_rel_path=smartlink(load_ete_tree_link, ete_tree_link)
      write('# <- link: {} to {}'.format(load_rel_path, ete_tree_link), how=how_found)
      os.symlink(load_rel_path, ete_tree_link)
    else:    
      build_cmd='ete3 build --noimg {x} -w {w} -a {a} -o {o} -v 1'.format(w=ete3_workflow, a=ali_w_context, o=ete_out_folder, x=ete_option)
      with open(ete_log, 'w') as fh:
        run(build_cmd, out=fh, err=fh, tolerate=True)

      last_lines_of_log=run('tail -n 30 {}'.format(ete_log))
      if not 'The following published software and/or methods were used.' in last_lines_of_log:
        raise Exception('ERROR ete3 failed, check: {}'.format(ete_log))
      
      dest_file=glob('{}/*/*.w_context.fa.final_tree.nw'.format(ete_out_folder))[0]
      link_cmd='ln -fs {} {}'.format(    smartlink(dest_file, ete_tree_link),     ete_tree_link)
      run(link_cmd)
  else:
    write('# <- found: {}'.format(ete_tree_link), how=how_found)

  write('#### routine: FIGURES      ######', how=how_routine)
  species_file=   '{}id2species'.format(output_prefix)
  desc_file=      '{}desc'.format(output_prefix)
  style_file=     '{}style'.format(output_prefix)
  figure_w_desc=  '{}tree.2.pdf'.format(output_prefix)
  figure_no_desc= '{}tree.pdf'.format(output_prefix)
  tax_file=       '{}tree.tax'.format(output_prefix)
  tax_temp=       '{}tree.tax.tmp'.format(output_prefix)
  
  with open(species_file, 'w') as fh:
    write(f'# writing -> {species_file}')
    fh.write('\n'.join( [s+'\t'+seqid2species[s]   for s in   seqid2species] ) +'\n' )
    fh.write('\n'.join( [title.split()[0]+'\tncbi' for title, seq in parse_fasta(context_sequences)]) +'\n')
  
  if not os.path.isfile(figure_w_desc) or opt['F']:
    desc_cmd='''gawk '/>/{{print substr($1, 2) "\\t" substr($0, length($1)+1, {n})}}' {i}'''.format(n=60, i=context_sequences)
    with open(desc_file, 'w') as fh:
      run(desc_cmd, out=fh, err=fh)
    style_cmd='''gawk '/>/{{print "i="substr($1, 2)":c={c}"}}' {i} '''.format(c='lightblue', i=input_file)
    with open(style_file, 'w') as fh:
      run(style_cmd, out=fh, err=fh)
    show_1_cmd='''show_tree.py -f {t} -y {s} -C -lf ' "a" ' -no_id -nb 1 -k {k} -st {st} -g -out {o}  '''.format(t=ete_tree_link, s=style_file, k=2, o=figure_no_desc, st=species_file)
    if not opt['nf']:

      try:    run(show_1_cmd)
      except: write('# ERROR show_tree.py did not work! Is a display available? -- attempting to continue')
      if not os.path.isfile(tax_file) and not opt['dr']:
        write("# weird error with getting tax. going one by one... ")
        tax_remedy_cmd='''for x in $(gawk '/>/{{print substr($1, 2)}}' {c} ); do fetch_ncbi_sequences.py -I $x -m P -t -M -L | gawk -F"\t" '{{$1=substr($1, 2); split($2, A, " -> "); $2=A[2]; $3=A[1]; OFS="\t"; print}}'  ; done > {to}; for x in $(gawk '/>/{{print substr($1, 2)}}' {i} ); do s=$(echo $x | gawk -F. '{{print $4}}'); echo ${{x}}%%%${{s}}%%%$(ncbi_taxonomy.py -S $s) | sed 's/%%%/\\t/g' >> {to}; done'''.format(c=context_sequences, i=input_file, to=tax_temp)        
        run(tax_remedy_cmd, shell=True)
        run('mv {to} {o}'.format(o=tax_file, to=tax_temp) )
        write("# repeating show tree command ...")
        run(show_1_cmd)
    else: write(show_1_cmd)       
        
    show_2_cmd='''show_tree.py -f {t} -y {s} -C -lf ' "a" ' -no_id -nb 1 -k {k} -st {st} -g -out {o} {d}'''.format(t=ete_tree_link, s=style_file, k=2, d=desc_file, o=figure_w_desc, st=species_file)
    if not opt['nf']:
      try:    run(show_2_cmd)
      except: write('# ERROR show_tree.py did not work! Is a display available? -- attempting to continue')
    else: write(show_2_cmd)

    





  write('#   end date: {}'.format(bbash('date'))) #, how=how_main)


  ###############




def smartlink(source, dest):
  """ Return the path that should be used to link source to dest accounting for relative paths """
  abs_source=os.path.abspath(source)
  abs_source_folder= '/'.join( abs_source.split('/')[:-1] )  # no trailing /
  if abs_source_folder.count('/'): abs_source_folder+='/'
  abs_dest=os.path.abspath(dest)
  abs_dest_folder=   '/'.join( abs_dest.split('/')[:-1]   )
  if abs_dest_folder.count('/'): abs_dest_folder+='/'

  chari=0  # to deal with limit case, no folder
  for chari, char in enumerate(abs_source_folder):
    if  chari+1>len(abs_dest_folder) or abs_dest_folder[chari]!=char: break
    chari+=1 # only to deal with case in which it exists without breaking
  #chari now marks the limit of first non-equal character
  common_path=abs_source_folder[:chari]
  rest_source=abs_source[chari:]
  rest_dest_folder=abs_dest_folder[chari:]
  n_back=rest_dest_folder.count('/')

  # print ( ('abs_source_folder', abs_source_folder) )
  # print ( ('abs_dest_folder', abs_dest_folder) )
  # print ( ('common', common_path) )
  # print ( ('rest_source', rest_source) )
  # print ( ('rest_dest_folder', rest_dest_folder) )

  dest_link_to='../'*n_back + rest_source  if common_path else  source

  # to deal with absolute inputs with nothing in common (algorithm would try to do relative paths until the very root; working but not pretty:
  if source in dest_link_to: dest_link_to=source  
  return (dest_link_to)


  

#######################################################################################################################################

def close_program():
  if 'temp_folder' in globals() and os.path.isdir(temp_folder):
    printerr(f'Program finished. Make sure you delete this: {temp_folder}')
    #shutil.rmtree(temp_folder)
    try:
      flush_service()
    except:
      pass


if __name__ == "__main__":
  try:
    main()
    close_program()  
  except Exception:
    close_program()
    raise 



#!/usr/bin/env python3
import sys
from ete3 import *
from .MMlib3 import *
from ncbi_db import ncbi_sequences
from ncbi_db import ncbi_taxonomy_tree
from ncbi_db import ncbi_taxonomy
from .tree_classes import *
import brewer2mpl ## nice colors
from math import log10
import random

class TextFace(faces.TextFace):
  def __init__(self,  *args, **kargs):
    if 'ftype' in kargs: pass
    else:                kargs['ftype']=opt['ft']
    return faces.TextFace.__init__(self,  *args, **kargs)

global help_msg
help_msg="""Program to show a tree of sequences using ete3. Accepts various id formats and provide options to customize the graphics.

Usage:    show_tree.py -f input_tree.newick [other_options] [feature_files] 

This program is thought for visualization of tree of sequences. It derives the species to which every node in the input tree belongs to, and their complete ncbi taxonomy. By default, node names can be gi codes from ncbi (like gi|12356715|whatever) which are interpreted as protein ids and fetched connecting to ncbi. Use option -nt to specify that the gi refers to a nucleotide entry. Node names can be selenoprofiles joined ids, like this: fam.1.label.species.target. Alternatively, other options are available to control species and taxonomy (see -h full).

## Tree attributes
-format forces ete3 to read the input tree in a certain newick specification. Def:0
-b      show branch support as values in the bottom of branch nodes
-s      do not show the column with the species name
-l      do not show the column with the taxonomy of the species. Taxonomy is still computed though: use -CL -1 to avoid fetching taxonomy from ncbi
-C      circular tree
-scale  defaults to 10, set higher to increase the size of the branches. when -C is active it gets *10
-ali    fasta alignment used to display the sequences directly in the nodes. The program tries also to derive the genomic coordinates of each result from the titles, to display intron positions. This will work only if titles are MMlib headers (e.g. selenoprofiles output). Otherwise, only the alignment structure is displayed and a warning is printed.

By default, the ETE interactive environment is opened to display the tree according to options. When it is closed, the program quits. 
Alternatively, one can:

-p      open interactive mode (prompt) after showing tree
-out    print file in output. Can be .png or .pdf ; if this option is active, the graphical environment will not be opened.
 -img_h  height in pixels of output file (only if -out is active)
 -img_w  width in pixels of output file  (only if -out is active)

To further customize visualization, see the advanced help page using  show_tree.py   -h full  """

advanced_help= """\n####### Advanced help page #######
-v            verbose mode. prints lots of stuff
-config       load options from this configuration file
-save_config  save options to this configuration file
-dry          do not produce any graphics (does not require X11)

## Setting outgroup; to call the duplications/losses events, the tree must be rooted. There are three ways to do this:
        by default, the outgroup is set using the ete3 Tree class method named get_midpoint_outgroup, which splits the tree according to branch length
-o      directly choose outgroup; the argument must be a node name. Otherwise, for ancestral nodes, you can define the outgroup node as last common ancestor of two or more nodes using the "&" symbol. E.g. "-o   'node_id1&node_id5'"
-age    species2age file (each line is species TAB age). if provided, the function get_age_balanced_outgroup is used to decide the outgroup. 

## Controlling size of elements added with option -ali
-ali_w      total width of face displayed (including gaps)
-ali_wpp    width per position (only if coordinates of the gene are recognized)
-ali_h      total height of face displayed (including gaps)
-ali_p      what is printed on this face; defaults to nothing. You can use comma separated arguments, e.g. chromosome,id

## Controlling style: node names, species and taxonomy displayed
-no_id  do not show face with node name
-nb     do not show colored balls next to nodes based on labels
-sf     species function. specify a python function based on "x" as the node name. Example:   -sf " x.split('.')[1] "; the result look like: "Drosophila melanogaster"; it can return 'ncbi', in which case it is taken as NCBI protein id
-st     species tab file (alternative to -sf). Each line must have this structure:    node_name TAB species name  (no space in between)
-um     unmask characters. Use with options -sf or -st if the species names are coded like: "Drosophila_melanogaster", "Xenopus_{ch40}Silurana{ch41}_tropicalis"
-umn    unmask ncbi ids before fetching. Use those ids with 'ncbi' assigned with -sf and -st are masked, e.g. XP{ch95}006124501{ch46}1
-cl     custom lineage summary. A tab separated file must be provided, with first field: species name (unmasked), and second field: custom lineage. The standard lineage fetched from the ncbi will be printed for species not found in the file
-CL     same as -cl, but it forces all names from the file provided. If no taxonomy is found for a species, it is left blank
-cs     common species translation; a tab separated file must be provided as argument
-tax    species and taxonomy information file. It is a tab separated file like: nodename TAB speciesname TAB taxonomyinfo. Actually one such .tax file is written at the end of computation, to derive all this information just once per tree.
-ids    show custom defined ids (instead of standard node names from input tree) next to each node. You must provide a tab separated file as argument
-g      tolerate errors in taxonomy determination (useful for big trees)
-mc     maximum number of characters in a species name
-ft     font type used in displayed text

## Duplications/speciations
-mfd    saturation number of species supporting a duplication. Even if an event has more support that this, it will not be displayed bigger

## Manipulating tree
-exec     execute some python commands given as string before calling t.show()
-remove   file with node names to be removed from the tree before displaying
-zoom     show only a portion of the tree, defined as the last common ancestor of two nodes. Argument= nodeName1&nodeName2
-log      transform branch lengths so that newDist=log10(1+dist)

## Controlling labels; these are categorical attributes of nodes which are color coded in the graphics'
-lab      label file, controlling each color and possibly an image to display next to each node. Format for each line: label -tab- color    OR   label -tab- color -tab- path_to_image
-lf       function to derive label from node name. specify a python function based on "x" as the node name. Example:   -sf " x.split('.')[2] "
-L        alternative to -lf to define label for each node. File format:  protein_id1 -tab- label
-lmap     if not -lab, then colors are automatically set; use this to define the options of the color settings with brewer2mpl.
          Format class:name  e.g. Qualitative:Set1

## Additional coloring:
-k      color scheme for species/lineage. Default ("1") is color by kingdom; "2" is color by eukaryotic lineage; "0" not to color
        do not color species and taxonomy according to kingdom.  default is Bacteria: red; Archaea: blue; Eukaryotes: green (non metazoa: darker green) 
        Another format for argument: lineage1:color1,lineage2:color2  etcetera
-y      file with style specifications, allowing to color backgrounds or text of certain nodes. 
 Format of each line:  which_nodes:what_style    where:
  which_nodes can be  i=NODEID or s=SPECIES or t=TAXONOMYTAG  or any comma-separated combination; ID1&ID2 can be used to call an ancestor
  what_style  can be  fgcolor=COLOR or bgcolor=COLOR (any attribute of Node.img_style in format WHAT=VALUE) or any comma-separated combination. c= is short for bgcolor=
  Examples of lines:  s=Homo_sapiens,s=Gorilla_gorilla:c=orange     t=Rodentia:fgcolor=red 
-Y      Like -y but provided on the command line instead; different 'lines' are here separated by ";" characters
-ac [n]   automatic color by duplication cluster. Experimental. Use -aco [ofile] to write file

## Features files; these are files to add custom columns for certain or all nodes. 
You can specify any feature files on the command line (see Usage above). These must have the following format; all delimiters shown from here below with more than one space are actually tab separators. Those shown as single spaces must be single spaces. Simplest example:

protein_id1    labelA
protein_id2    labelB
protein_idN    labelZ

Additional information can be included in the header of the feature file, to define the color with which the feature label will be printed. If undefined, black is used.
#FEATURE_TITLE cysteine #880000   selenocysteine #008800
Aan.C3    cysteine
Aan.U13   selenocysteine
Aan.U13a  selenocysteine

"""
def verbose(msg):
  global opt
  if opt['v']:   write(msg, 1)
  else:          service(msg)

def label_to_image(label):
  """This function is used when you define a particular feature file with option -L ; each node is drawed using a colored ball. The image for the ball is searched into image_folder, the file name depends on the label translated through label2image . If the label is not present among the keys, the image corresponding to the label "unknown" is used. """
  if label in label2image:    return label2image[label]
  else:                             return None

def label_to_color(label):
  if label in label2color:    return label2color[label]
  else:                             return label2color['unknown']
  
class BallImgFace(faces.ImgFace):
  def update_pixmap(self):
    self.pixmap = faces.QPixmap(self.img_file).scaled(15, 15, faces.Qt.KeepAspectRatio)

def get_duplication_size(node):
  """ function that assigns the size of a duplication node for displaying, in order to display nodes with higher reliability, given by the number of species supporting the duplication"""
  global opt
  if len(node.children)!=2 : return 4
  a, b = node.children
  set_a, set_b = set(), set()
  for n in a.traverse(): 
    if n.is_leaf(): set_a.add(protein2species[n.name])
  for n in b.traverse(): 
    if n.is_leaf():       set_b.add(protein2species[n.name])
  n_species= len(set_a.intersection(set_b))
  size= min(2+ n_species, 2+opt['mfd']) #more than 12 species will be displayed as 
  return size

def is_selenoprofiles_title(short_title): 
  short_title=short_title.split()[0]
  return short_title.count('.')==4 and is_number(  short_title.split('.')[1] )

lineage_color_specs=None
def get_color_for_lineage(t, s, option):
  t+='; '
  global lineage_color_specs
  if option==1:
    if   'Bacteria;'  in t: c="#990000"     
    elif 'Archaea;'   in t: c="#000099"     
    elif 'Eukaryota;' in t:  
      #drawing a slighly different color for non-metazoa
      if 'Metazoa; '  in t:          c="#009900"     
      else:                          c="#004400"     
    else:       c='#000000'
  elif option==2:
    if   not 'Metazoa; ' in t:    
      if   'Fungi; ' in t:          c='orchid'
      elif 'Chlorophyta; ' in t:    c='#222222'
      elif 'Archaea; ' in t:        c='#0A0A0A'
      elif 'Bacteria; ' in t:       c='#131313'
      else:                       c='#727073'
    elif not 'Bilateria; ' in t:  c='#ACAAAD'
    else: 
      if 'Deuterostomia; '   in t:
        if 'Vertebrata; '    in t:     
          if   'Mammalia; '    in t:       
            if   'Primates; ' in t:              c='#009900' #green            
            elif 'Rodentia; ' in t:              c='#00AA33' #green
            else:                                c='#187700' #green 
          elif 'Sauropsida; '  in t:       c='#004400' #darkergreen
          elif 'Amphibia; '    in t:       c='#178A71' #cyan
          elif 'Actinopterygii; '   in t:  c='#249BC9' #blue
          else:                            c='#153EB0'  #other vertebrates  #darkerblue
        else: c='#382AB8' #other deuteros    #darkpurple
      elif 'Protostomia; '    in t:
        if   'Insecta; '    in t:   c='#D10000'  #red
        elif 'Arthropoda; ' in t:   c='#690707'  #darker red    
        elif 'Nematoda; '    in t:  c='#E016A4'  #violet
        elif 'Mollusca; '    in t:  c='#D18C0A'  #orange
        else:                       c='#B0B518'  #yellow
      else:
          c='#000000'  #any bilaterian is not protos and not deuteros ?     
  elif option==0:  c="#004400"
  else:
    if lineage_color_specs is None:
      lineage_color_specs=[]
      for piece in option.split(','):
        lineage, color=piece.split(':')
        lineage_color_specs.append([lineage,color])
    c='#000000'
    for lineage, color in lineage_color_specs:
      if lineage+'; ' in t or s.startswith(lineage): c=color
  return c 

## Function to draw the tree. This is applied to every node
def mylayout(node):
  ## Phylogeny layout
  if hasattr(node,"evoltype"):
    if node.evoltype == 'D':
      node.img_style["fgcolor"] = "#1111FF"
      node.img_style["size"] = get_duplication_size(node)
    elif node.evoltype == 'S':
      node.img_style["fgcolor"] = "#FF1111"
      #node.img_style["line_color"] = "#FF0000"
      node.img_style["size"] = 4
  if not node.is_leaf():
    if node.name in style_dictionary['id']:     
      for attribute in style_dictionary['id'][node.name]:
        node.img_style[attribute] = style_dictionary['id'][node.name][attribute]
  else:
    node.img_style["size"] = 0
    node.img_style["fgcolor"] = "#000000"
    column_index=0
    label=id2label[node.name]
    label_color=label_to_color( label )
    image_path=label_to_image(  label  )
    if not image_path is None and not opt['nb']:
      ballFace=BallImgFace(image_path)
      faces.add_face_to_node(ballFace, node, column = column_index, aligned = False)
      column_index+=1
    elif image_path is None and not opt['nb']:         #not
      circleFace= faces.CircleFace( 4, label_color, style='sphere', label=None)
      faces.add_face_to_node(circleFace, node, column = column_index, aligned = False)
      column_index+=1
    else:
      column_index+=1

    # id writing
    if not opt['no_id']:
      fg_color="#000000"
      node_name_to_write=node.name
      if node_name_to_write.startswith('gi|'): node_name_to_write='gi|'+node_name_to_write.split('|')[1]
      elif is_selenoprofiles_title(node.name): node_name_to_write= node.name.split('.')[1]
      if node.name in names_to_show:           node_name_to_write= names_to_show[node.name]   ### if opt['ids'] is active
      nameFace=TextFace(node_name_to_write, fgcolor=fg_color)
      nameFace.margin_left=1; nameFace.margin_right=4
      faces.add_face_to_node(nameFace, node, column = column_index, aligned = False)
      column_index+=1

    for face in node.extra_faces:
      faces.add_face_to_node(face, node, column = column_index, aligned = True)
      column_index+=1

    if True: ###not (opt['s'] and not opt['k'] and opt['l']):
      #drawing species column
      species_name= protein2species[node.name]
      text_color="#000000"
      tax_string=''
      if opt['k'] or not opt['l']:
        #computing full lineage
        if opt['CL']:        tax_string=''
        else:                
          try:          tax_string=taxonomy_lineages[species_name]
          except KeyError: 
            if not opt['g']: raise
            printerr('WARNING no taxonomy for this species: {s}'.format(s=species_name), 1)
            tax_string=''
      if opt['k']:
        #setting color
        text_color=get_color_for_lineage(tax_string, species_name, opt['k'])

      if not opt['s']:    
        #adding species column
        verbose( "drawing "+ node.name +"\t\tspecies:"+ species_name)
        species_name_to_show= species_name
        if len(species_name_to_show)>opt['mc']: species_name_to_show=species_name_to_show[:opt['mc']]+'*'
        if species_name in species_common_names:          species_name_to_show=species_common_names[species_name]    ### for opt['cs']
        sp_face=TextFace(species_name_to_show+' ' , fgcolor=text_color) #######
        sp_face.margin_left=5;         sp_face.margin_right=5
        faces.add_face_to_node(sp_face, node, column = column_index, aligned = True)
        column_index+=1

      if not opt['l']:
        #drawing lineage column
        summ_lineage=lineage_string_to_abstract(tax_string)
        if species_name in custom_lineages:   
          summ_lineage= custom_lineages[species_name]
          write( (tax_string, summ_lineage, species_name), 1)

        lineage_face=TextFace(summ_lineage+' ' , fgcolor=text_color) 
        faces.add_face_to_node(lineage_face, node, column = column_index, aligned = True)
        column_index+=1

      pos_dict={}   ### all this block does nothing unless there's stuff in style_dictionary['tax']
      for taxtag in style_dictionary['tax']:
        f=tax_string.find ( '{}; '.format(taxtag) )
        if f!=-1: pos_dict[taxtag]=f
      for taxtag in sorted( list(pos_dict.keys()), key=lambda x:pos_dict[x] ):
        for attribute in style_dictionary['tax'][taxtag]:
          node.img_style[attribute] = style_dictionary['tax'][taxtag][attribute]         

      if species_name in style_dictionary['species']:
        for attribute in style_dictionary['species'][species_name]:
          node.img_style[attribute] = style_dictionary['species'][species_name][attribute]

      if node.name in style_dictionary['id']:     
        for attribute in style_dictionary['id'][node.name]:
          node.img_style[attribute] = style_dictionary['id'][node.name][attribute]
    
    """
    if opt['seq']:
      if hasattr(node, 'sequence') and node.sequence:
        seq_face= faces.SequenceFace( node.sequence, 'nt')
        faces.add_face_to_node(seq_face, node, column = column_index, aligned = True)
        column_index+=1"""

    for feature_index in range(len(features_dictionaries)):
      features_dict= features_dictionaries[feature_index]
      features_dict_label_to_color= features_dictionaries_label_to_color[feature_index]
      text='';       color='#000000'
      if node.name in features_dict: text= features_dict[node.name]
      if text in features_dict_label_to_color: color=features_dict_label_to_color[text]
      feature_face=TextFace(text, fgcolor=color)
      faces.add_face_to_node(feature_face, node, column = column_index, aligned = True)
      column_index+=1

#####################################################################
###########################  PROGRAM START ##########################
## ***** ***** ***** ***** ***** ***** ***** ***** ***** ***** *** ##
#####################################################################

protein2species={}              #filled with names during the execution of the program
taxonomy_lineages={}         #filed with key: species_name -> full lineage "; separated)
style_dictionary={'id':{}, 'species':{}, 'tax':{}}   ###### 

def main(args={}):
  command_line_synonyms={'t':'f', 'i':'f', 'style':'y', 'Style':'Y'}
  def_opt= {'f':'tree_inpufile.newick',  'o':'', 'C':'', 'S':0, 's':0, 'k':1, 'l':0, 'L':0,'*':[], 'p':0, 'out':'', 'ali':'', 'bsize':8, 'T':1, 'n':0, 'b':0, 'cs':0, 'ids':0, 'width':150, 'sf':None, 'st':None, 'um':0,  'tax':None, 'cl':0, 'CL':0, 'nt':0, 'ft':'Helvetica',
            'lab':0,
            'no_id':0, 'lf':0, 'scale':10, 'format':0, 'g':0, 'mc':35,  'mfd':12, 'nb':0, 'y':0, 'Y':0,
            'umn':0,
            'get_species':None, 'remove':0, 'age':0, 'img_h':None, 'img_w':None, 'v':0, 'temp':'/tmp/', 'exec':0, 
            'ali_w':150,  'ali_wpp':None, 'ali_h':10,    'ali_p':0, 
            'config':0, 'save_config':0, 'zoom':0, 'lmap':'Qualitative:Paired', 'log':0,   'ac':0, 'aco':0,
            'dry':0,
}

  config_filename=def_opt['config']
  for i in range(len(sys.argv)):
    if sys.argv[i] == "-config":     
      config_filename=sys.argv[i+1]
  if config_filename:
    def_opt.update( configuration_file(config_filename) )
    def_opt['*']=eval(def_opt['*'])
  global opt
  global all_seqs_alignment, profile_ali
  global label2image
  global label2color
  if not args: opt=command_line(def_opt, help_msg, '*', synonyms=command_line_synonyms, advanced={'full':advanced_help} )
  else:  opt=args
  set_MMlib_var('opt', opt)
  #########################################################
  ############ loading options
  if opt['save_config']:
    config_text='\n'.join([ '{} = {}'.format(k, opt[k]) for k in sorted(opt.keys()) if not (k.startswith('__') or k in set(['save_config', 'config'])   )] )
    write('Saving config --> {}'.format(opt['save_config']), 1)
    write_to_file(config_text, opt['save_config'])

  global temp_folder; temp_folder=Folder(random_folder(opt['temp'])); test_writeable_folder(temp_folder, 'temp_folder'); set_MMlib_var('temp_folder', temp_folder)
  inputfile=opt['f']; check_file_presence(inputfile, 'newick inputfile')

  ##### LOADING TREE !
  t = PhyloTree(inputfile, format=opt['format'])
  t.set_species_naming_function(  lambda x:protein2species[ x ] )
  for node in t: node.extra_faces=[]

  ############################### 
  ##### setting outgroup  
  if opt['o']:
    try:
      if "&" in opt['o']:
        outgroup= t.get_common_ancestor(   opt['o'].split('&') )   
      else:
        matches = t.search_nodes(name = opt['o'])
        outgroup=matches[0]
      write('OUTGROUP: setting with option -o -> '+opt['o'], 1)
      t.set_outgroup(outgroup);
    except:            raise Exception("ERROR setting outgroup with option -o, the node was not found: "+str(opt['o']))
  elif opt['age']:
    species2age={};       unmasked_species2age={}
    for line in open(opt['age']): 
      species2age[line.split('\t')[0]]= int( line.strip().split('\t')[1] )
      unmasked_species2age[     unmask_characters(  line.split('\t')[0].replace( '_', ' '))    ] = int( line.strip().split('\t')[1] )
    write('OUTGROUP: setting with get_age_balanced_outgroup method ... ', 1)
    try:                 outgroup=t.get_age_balanced_outgroup(species2age)
    except KeyError: 
      try:               outgroup=t.get_age_balanced_outgroup(unmasked_species2age)
      except KeyError:   
        printerr( "ERROR species not found in -age file "+opt['age'] +' ! ', 1)
        raise 
    t.set_outgroup(outgroup)
  else:
    write('OUTGROUP: setting with get_midpoint_outgroup method ... ', 1)      
    t.set_outgroup(t.get_midpoint_outgroup())
  t.ladderize()

  if opt['style'] or opt['Style']:
    unnamed_index=0
    write('Reading style specifications...', 1)
    style_text=opt['Style'] if not opt['style'] else ';'.join([line.strip() for line in open(opt['style'])])
    for piece in style_text.split(';'):      
      if not piece.strip(): continue
      which_nodes, what_style=piece.strip().split(':')
      stylespec={}
      for p in what_style.split(','):
        attribute,value=p.strip().split('=')
        if attribute=='c': attribute='bgcolor'
        value=option_value(value) #converting to int/float useful for certain attributes
        stylespec[attribute]=value
      for p in which_nodes.split(','):
        clas,value=p.strip().split('=')
        if   clas=='i': 
          if '&' in value: 
            id1,id2=value.split('&')
            anc=t.get_common_ancestor(id1, id2)
            if not anc.name: 
              anc.name='Node{}'.format(unnamed_index)
              unnamed_index+=1
            value=anc.name            
          style_dictionary['id'][value]=stylespec
        elif clas=='s': style_dictionary['species'][value]=stylespec
        elif clas=='t': style_dictionary['tax'][value]=stylespec
        else:     raise Exception("ERROR reading style failed {}  --> first bit of which_nodes part should be one of these: i s t  while it is: {}".format(style_text, clas))

  ### removing nodes if option is active
  if opt['remove']:
    write('Removing some leaf nodes from the tree... ', 1)
    ids_to_remove={}
    for line in open(opt['remove']):   ids_to_remove[line.strip()]=1
    nodes_to_keep=[]
    for node in t.traverse():
      if node.is_leaf():
        if node.name in ids_to_remove:  ids_to_remove[node.name]=0
        else: nodes_to_keep.append(node.name)
    if any( ids_to_remove.values() ):
      printerr('WARNING some nodes that were supposed to be removed were not found:\n'+join(  [an_id for an_id in ids_to_remove if  ids_to_remove[an_id] ] , '\n' ), 1   )
    t.prune(nodes_to_keep)

  #####################################################################
  #####################################################################
  #########  determining species for each node in the input protein tree
  ### checking if already computed or provided with option -tax
  if is_file(inputfile+'.tax') and not opt['tax']:  taxonomy_tab_file= inputfile+'.tax'
  else:                                             taxonomy_tab_file= opt['tax']
  if taxonomy_tab_file:
    try:
      for line in open(taxonomy_tab_file, 'r'):
        node_name, species_name, taxonomy_lineage = line[:-1].split('\t')
        protein2species[node_name]=species_name
        taxonomy_lineages[species_name]= taxonomy_lineage
    except:       printerr('ERROR parsing taxonomy file! \n{}'.format(line), 1);      raise
    for node in t:
      try: 
        if node.name not in protein2species:               
          raise Exception('ERROR species not found for node: '+node.name+' in -tax file: '+taxonomy_tab_file)
        if protein2species[node.name] not in taxonomy_lineages: raise Exception('ERROR taxonomy not found for species: '+protein2species[node.name]+' in -tax file: '+taxonomy_tab_file)
      except: 
        if not opt['g']: raise
        printerr('WARNING something missing in file {f} for node {s}'.format(f=taxonomy_tab_file, s=node.name))
        protein2species[node.name]=node.name
        taxonomy_lineages[  protein2species[node.name]  ] = 'unknown'

  else:
    ncbi_names_to_fetch={} # name_to_fetch -2-> node_name
    if not opt['sf'] is None:
      if opt['sf'].lower()=='ncbi':  species_function= lambda x:"ncbi"
      else:                         species_function=eval('lambda x:'+opt['sf'])
      for node in t:
        try:
          species= species_function(node.name)
          if opt['um']:   species= unmask_characters( species.replace( '_', ' ') )
          assert species
          if species.lower()=='ncbi':
            ncbi_names_to_fetch [node.name] = node.name
          else:
            protein2species[node.name] = species
        except: 
          printerr('ERROR species inferring function (-sf) not working on node: '+node.name, 1)
          raise
    elif not opt['st'] is None:
      for line in open(opt['st']):
        if line.startswith('#'): continue
        splt=line.strip().split('\t')
        protein_name=splt[0]; species=splt[1]
        if opt['um']:   species= unmask_characters( species.replace('_', ' ') )
        if species=='ncbi':
          ncbi_names_to_fetch [protein_name] = protein_name
        else:
          protein2species[protein_name] = species

        
      for node in t: 
        if node.name not in protein2species and node.name not in ncbi_names_to_fetch: 
          raise Exception('ERROR species not found for node: '+node.name+' in -st file: '+opt['st'])
    else:
      #### no species explicitly chosen. Trying to guess, if the title is a selenoprofiles title or a ncbi gi code
      for node in t:
        if is_selenoprofiles_title(node.name):
          masked_species= node.name.split('.')[3]
          species= unmask_characters( masked_species.replace(  '_', ' ') )
          protein2species[node.name] = species
        elif len(node.name)>3 and node.name[:3]=='gi|':    
          id_to_fetch='gi|'+node.name.split('|')[1]
          ncbi_names_to_fetch [id_to_fetch] = node.name
        else: 
          raise Exception("ERROR no species could be parsed from protein title:" +node.name)

    ## fetchin gi codes all at once
    if opt['umn']:
      ncbi_names_to_fetch={ unmask_characters(k):ncbi_names_to_fetch[k]    for k in ncbi_names_to_fetch}
    if ncbi_names_to_fetch and not opt['CL']:
        verbose('fetching species names for '+str(len(ncbi_names_to_fetch))+' proteins ... ')
        sequence_type_mode='P'
        if opt['nt']: sequence_type_mode='N'
        # write('fetching now', 1)
        # try:
        #write('\n'.join( sorted(ncbi_names_to_fetch.keys()) ) , 1)
        #print(ncbi_names_to_fetch.keys())
        fetch_output= fetch_ncbi_sequences.main({'I':   ','.join(list(ncbi_names_to_fetch.keys())), 'silent':1, 't':1, 'm':sequence_type_mode}) # #hash. 
        #   write('finished fetching', 1)
        #   write('\n'.join( [(a,b) for a,b in fetch_output.iteritems()]), 1)
        # except:
        #   raise
        
        for gi_code in list(ncbi_names_to_fetch.keys()):
          if not gi_code in fetch_output: raise Exception("ERROR could not obtain from ncbi the species from code: "+gi_code+' ; protein name: '+ncbi_names_to_fetch[gi_code])
          try: 
            species=   fetch_output[gi_code][1].split(' -> ')[-1]
            taxonomy=  fetch_output[gi_code][1].split(' -> ')[0]
          except:
            printerr('WARNING: could not fetch species / taxonomy for this entry: {}\n{}'.format(gi_code, fetch_output[gi_code][1]), 1)
            species, taxonomy='unknown', 'unknown'
                    
          protein2species[ncbi_names_to_fetch[gi_code]] = species
          if species not in taxonomy_lineages: taxonomy_lineages[species]=taxonomy

    if not opt['CL']:
      ### getting full taxonomy lineage for all species. First we try with ncbi_taxonomy_tree ; this is local and fasta but sometimes misses some entries, and we use ncbi_taxonomy (uses internet) for the rest
      species_to_fetch= sorted(list(set([ species for species in list(protein2species.values())         if species not in taxonomy_lineages ])))
      if species_to_fetch:
        write_to_file( '\n'.join(species_to_fetch), temp_folder+'species_to_fetch')
        #write(' debug: '+ temp_folder+'species_to_fetch', 1)
        verbose('getting taxonomy lineage for all species... ')
        try:
          args = ncbi_taxonomy_tree.parse_opts( '-n {0} -u --lineage --quiet -f -temp {1} '.format(temp_folder+'species_to_fetch', temp_folder) )
          lineage_annotated_tree =  ncbi_taxonomy_tree.main(args) #  TEMPORARY NO NCBI TAX TREE
          #print( lineage_annotated_tree )
        except:
          input('ERROR with ncbi_taxonomy_tree.py -n {0} -u --lineage --quiet -f -temp {1} '.format(temp_folder+'species_to_fetch', temp_folder)+ ' ... press enter to continue')
          raise 
        missing_species={}
        for species in species_to_fetch:
          try:
            unmasked_species= unmask_characters(species.replace('_', ' '))
            leaf = lineage_annotated_tree & unmasked_species
            taxonomy = '; '.join(leaf.lineage.split('||')[:-1])  #replacing || with "; " and taking off last field (species name)
            taxonomy_lineages[species]=taxonomy
          except: 
            missing_species[species]=1
        ## some were missed! fecthing online
        for species in missing_species:
          verbose('getting taxonomy lineage for missing species: '+species)
          tax_results = ncbi_taxonomy.main(  {'silent':1, 'S':species}  )
          if not len(tax_results): 
            if opt['g']: 
              printerr("WARNING can't find taxonomy entry for species: "+species, 1)  # Exception,  "ERROR can't find taxonomy entry for species: "+species              
              taxonomy_lineages[species]='unknown taxonomy'

            else: 
              #printerr("ERROR can't find taxonomy entry for species: "+species, 1)  # 
              raise Exception("ERROR can't find taxonomy entry for species: "+species)

          else:
            #print species, tax_results
            taxonomy=tax_results[list(tax_results.keys())[0]].split(': ')[1].split(' ->')[0]
            taxonomy_lineages[species]=taxonomy

      try:
          taxonomy_feature_file=inputfile+'.tax'
          taxonomy_feature_file_h=open(taxonomy_feature_file, 'w')
          for node in t:
            species_name=protein2species[node.name]
            print(node.name+'\t'+species_name+'\t'+taxonomy_lineages[species_name], file=taxonomy_feature_file_h)
          taxonomy_feature_file_h.close()
      except:      
        printerr('WARNING could not write taxonomy information file in '+taxonomy_feature_file+' ; the species and taxonomic lineages will be recomputed next time!', 1  )
        bash('rm '+taxonomy_feature_file)

  ##################### reading alignment file to be linked with the protein tree
  if opt['gff']:
    pass

  ########### labels file!
  label2image={}; label2color={}
  if opt['lab']:
    check_file_presence(opt['lab'], '-lab labels file')
    for line in open(opt['lab']):
      if not line[:-1]: continue
      splt=line.strip().split('\t')
      label2color[splt[0]]=splt[1]
      if len(splt)>2:
        label2image[splt[0]]=splt[2]    

  if not 'unknown' in  label2color:
    printerr('WARNING no default defined for "unknown" label! Please edit your -lab file', 1)
    label2color['unknown']='grey'
  ############ labels !  filling id2label
  global id2label; id2label={}
  if opt['L']:
    check_file_presence(opt['L'], 'labels feature file'); feature_file_h=open(opt['L'], 'r')
    line=feature_file_h.readline()
    while line and line[0]=='#':     line=feature_file_h.readline()
    while line:
      code, label=line[:-1].split('\t') #parsing feature file. code and label must be tab separated
      id2label[code]=label
      line=feature_file_h.readline()
    feature_file_h.close()
  else:
    if opt['lf']:
      try:      label_function=eval('lambda x:'+opt['lf'])
      except:   raise Exception("ERROR label function not valid: "+opt['lf'])
    else: 
      label_function=  lambda x:x.split()[0].split('.')[-3] ###trying selenoprofiles   lf
    for node in t:
      try:        
        label=label_function(  node.name ) or 'unknown'
        #print "LABELING", node.name, label
      except:     
        printerr("WARNING label function cannot determine label for node: "+node.name, 1)
        label='unknown'
      id2label[node.name]=label


  unassigned_labels=[label for label in sorted(set( id2label.values() ) )     if not label in label2color]
  map_type, map_name=opt['lmap'].split(':')
  colormap={}
  for strk in brewer2mpl.COLOR_MAPS[map_type][map_name]:
    colormap[int(strk)]=brewer2mpl.COLOR_MAPS[map_type][map_name][strk]
  
  if max(  colormap.keys()  )< len(unassigned_labels):
    printerr('WARNING too many labels with unassigned colors, skipping automatic label coloring', 1)
  else:    
    n_colors=len(unassigned_labels) if len(unassigned_labels) in colormap  else min( colormap.keys() )
    colormap=brewer2mpl.get_map(map_name, map_type, n_colors)
    for i, label in enumerate(unassigned_labels):
      label2color[label]=colormap.hex_colors[i]

  if opt['ali']:
    from PyQt5 import QtCore, QtGui
    #from selenoprofiles_tree_drawer import limited_p2ghit, reduce_font_if_necessary, GeneFace, set_selenoprofiles_tree_drawer_var    
    #set_selenoprofiles_tree_drawer_var('gene_brick_height', 15)
    #set_selenoprofiles_tree_drawer_var('gene_brick_width', opt['width'])    
    global_alignment=alignment(opt['ali'])
    #profile_ali=alignment()
    for title in global_alignment.titles():
      short_title=title.split()[0]
      aligned_seq=global_alignment.seq_of(title)
      try:        node=t&short_title  # --> searching the title in alignment in the tree
      except: 
        printerr( "-ali WARNING title: "+short_title+" not found in tree!", 1)
        continue
      label=id2label[short_title]
      label_color=label_to_color( label )
      color_bkg=label_color
      printed_hash= {'id':0, 'chromosome':0, 'boundaries':0, 'text':0}
      other_attributes={'height':opt['ali_h'], 'width':opt['ali_w'], 'width_per_position':opt['ali_wpp']}

      if opt['ali_p']:
        for field in opt['ali_p'].split(','):          printed_hash[field]=1


      try: # trying to infer gene structure from title        
        g=gene(); g.load_from_header(title)
        assert len(g.exons)
        if g.length() == (len(nogap(aligned_seq))+1)*3:  # sequence had the stop codon removed, trying to correct
            if   g.strand=='+':            g.exons[-1][1]-=3
            elif g.strand=='-':            g.exons[-1][0]+=3

        if g.length() != len(nogap(aligned_seq))*3:
          printerr('{} {} '.format(g.length(), len(nogap(aligned_seq))*3) , 1)
          printerr( "ERROR inferred gene length is not 3 times the length of sequence in the alignment! title: "+title+' ')

        exon_view_face=ExonViewFace(g, gaps=aligned_seq,     printed=printed_hash, colors={'bkg':color_bkg, 'outline':color_bkg}, **other_attributes)
        node.extra_faces.append(exon_view_face)
      except:
        for k in printed_hash: printed_hash[k]=0
        printerr( "-ali WARNING could not get gene coordinates from: "+title, 1)
        g=gene(chromosome='unknown', strand='+'); g.add_exon(1, len(nogap(aligned_seq)*3 ) )
        exon_view_face=ExonViewFace(g, gaps=aligned_seq,     printed=printed_hash, colors={'bkg':color_bkg, 'outline':color_bkg}, **other_attributes)
        node.extra_faces.append(exon_view_face)

    for node in t:
      if len(node.extra_faces) < 1:
        printerr( "-ali WARNING no alignment sequence found for tree node: "+node.name, 1)
        node.extra_faces.append( TextFace('') )



  """
  ### other sequences, these are dsplayed 
  if opt['seq']:
    seq_alignment=alignment(opt['seq'])
    t.link_to_alignment(alignment=seq_alignment.display(return_it=True), alg_format="fasta")
    for title in seq_alignment.titles():    
      try:
        n=t&title.split()[0]  # --> searching the title in alignment in the tree    
        n.sequence= seq_alignment.seq_of(title)
      except:        
        printerr('WARNING -seq : couldn\'t find title in alignment: '+title, 1)      """

  ####### loading custom ids
  global names_to_show; names_to_show={}
  if opt['ids']:
    for line in open(opt['ids']):      names_to_show[line.rstrip().split('\t')[0]] = line.rstrip().split('\t')[1]
  ####### loading custom species names
  global species_common_names; species_common_names={}
  if opt['cs']:
    for line in open(opt['cs']):      species_common_names[line.rstrip().split('\t')[0]] = line.rstrip().split('\t')[1]
  global custom_lineages;      custom_lineages={}
  if opt['cl'] or opt['CL']:    
    filename= opt['cl'] or opt['CL']
    if not opt['CL']==-1: 
      for line in open(filename):          custom_lineages[line.strip().split('\t')[0]]= line.strip().split('\t')[1]
    
  ################################## reading all features files
  global features_dictionaries;  features_dictionaries=[]
  global features_dictionaries_label_to_color; features_dictionaries_label_to_color=[]
  for feature_file in opt['*']:
    #parsing feature files
    check_file_presence(feature_file, 'feature_file'); feature_file_h=open(feature_file, 'r')
    code_to_label_diz={}; label_to_color_diz={}
    line=feature_file_h.readline()
    if line[0]=='#':
      #header; determining color for each label
      title=line.split('#')[1].split()[0]
      if title[0]==' ': title=feature_file
      for label_color in ' '.join(line.split(' ')[1:]).split('\t'): #parsing header of a feature file
        #print [label_color]
        label=label_color.split('#')[0][:-1] #parsing header of a feature file
        color='#'+label_color.split('#')[1]  #parsing header of a feature file
        label_to_color_diz[label]=color
      line=feature_file_h.readline() #jumping to first entry line, not commented.
    features_dictionaries_label_to_color.append(deepcopy(label_to_color_diz))
    #now parsing and adding one entry for each code found
    while line:
      code, label=line[:-1].split('\t') #parsing feature file. code and label must be tab separated
      code_to_label_diz[code]=label
      line=feature_file_h.readline()
    feature_file_h.close()
    features_dictionaries.append(deepcopy(code_to_label_diz))
    
  ########### executing commands
  if opt['exec']:
    try:
      exec(str(opt['exec']), globals(), locals())
    except Exception as e:
      write( "ERROR trying to evaluate: "+str(opt['exec'])+'; message:'+str(e), 1)
      raise

  if opt['log']:
    for n in t.traverse():            n.dist=log10(1+n.dist)

  ##########################
  ################### here we go! let's call evol events by species overlap
  ee=t.get_descendant_evol_events()
  #duplication_clusters(t)
  if opt['ac']:
    ## automatic clusters
    list_nodes2cluster, leaf2clustername=    make_clusters(t, species_fn=lambda x:protein2species[x.name], min_species_dup=opt['ac'], min_div_dist=2, min_species_div=sys.maxsize)
    if opt['aco']:
      ffw=open(opt['aco'], 'w')
      for leaf, cname in leaf2clustername.items(): print('{}\t{}'.format(leaf.name, cname), file=ffw)
      ffw.close()
    
    all_clusters=set(leaf2clustername.values())
    c1='#F0F0F0'; c2='#666666'    
    colors=[  color_scale( index/float(len(all_clusters)), c1, c2)   for index, cluster in enumerate(all_clusters)  ]
    random.shuffle(colors)
    cluster2color={ cluster: colors[index]  for index, cluster in enumerate(all_clusters)  }
    
    unnamed_index=-1
    
    for n,clust in list_nodes2cluster:
      if not n.name: 
        n.name='Node{}'.format(unnamed_index)
        unnamed_index-=1
      style_dictionary['id'][n.name]={'bgcolor':  cluster2color[clust] }

  
  tree_style=TreeStyle()
  tree_style.show_leaf_name=False
  tree_style.scale = opt['scale']
  if opt['b']: tree_style.show_branch_support=True  
  if opt['C']:
    tree_style.mode = "c" # draw tree in circular mode circular_style.scale = 20
    tree_style.allow_face_overlap = True
    tree_style.scale *= 10
    #tree_style.arc_start=270
  if opt['title']:
    tree_style.title.add_face(   TextFace(opt['title'], fgcolor='darkblue'), 0      )

  if opt['zoom']:   
    lca_nodes=opt['zoom'].split('&')
    t=t.get_common_ancestor(*lca_nodes)

  if opt['dry']:
    write('Dry run: not producing any graphics. Exiting... ', 1)
  else:
    if not opt['out']: t.show(mylayout, tree_style)
    else:              t.render(opt['out'], layout=mylayout, tree_style=tree_style, h=opt['img_h'], w=opt['img_w'])
  
  if opt['p']:     interactive_mode(message="The ete3.PhyloTree object is call t\nUse t.show(mylayout, tree_style) to display again the tree after your modifications")()

# ***** ***** ***** ***** ***** ***** ***** ***** *****
############################################################# PROGRAM END

def close_program():
  if 'temp_folder' in globals() and is_directory(temp_folder):
    bbash('rm -r '+temp_folder)
  try:
    if get_MMlib_var('printed_rchar'): 
      printerr('\r'+printed_rchar*' ' ) #flushing service msg space       
  except:
    pass

if __name__ == "__main__":
  try:
    main()
    close_program()  
  except Exception:
    close_program()
    raise 



#! /usr/bin/env python3
from string import *
import sys
from subprocess import *
import ete3  #from ete3 import * 
from math import log10
from .MMlib3 import  *
#write, printerr, write_to_file

help_msg="""Generic script to manipulate trees. 

Usage: $ tree_tools.py  -i tree.nw  [-o output_tree.nw]  [options]  

### Format:
-if      newick format of input (numeric, see ete3 help)
-of      newick format of output; if not provided, same as input
-off     features exported in output, comma separated; use 'all' to use them all

### Name manipulations:
-names   change node names from this file, which must be like:  name_in_tree TAB new_name
-mask    mask characters in node names  (see MMlib.mask_characters)
-unmask  unmask characters in node names 

### Structure manipulations:
-exp     expand certain leaves by adding a sister leaf (or a cluster) at distance 0. File must have lines like: leaf_in_tree TAB leaf_add1 [leaf_addN]
-um      make ultrametric; tree length is 1.0 or the argument provided; use -1 to use max leaf distance
-log     transform branch lengths so that newDist=log10(1+dist)
-std     standardize: nodes with only one child are removed and multifurcations are automatically resolved
-prn     prune: remove leaf nodes which are not in this file list
-lca     zoom into an internal node, defined as the last common ancestor of a list (>1) of nodes. Arg: nodeName1:nodeName2
-lad     ladderize tree
-mp      set outgroup using midpoint method

### Actions:
-print   print the tree with ete
-show    show the tree with ete; if a file is provided, render to that file instead
-o       output newick to this file, using the format in -of (see above)
-on      output names of leaves to this file, one per line
-P       open IPython interactive environment at the end

### Options:
-print_opt      print currently active options
-h OR --help    print this help and exit"""

command_line_synonyms={'p':'print'}

def_opt= {'temp':'/tmp/', 
          'i':0,  'if':0, 'off':'',
          'names':0, 'exp':0, 'mask':0, 'unmask':0, 
          'um':0, 'std':0, 'lad':0, 'log':0, 'prn':0, 'lca':0,
          'print':0, 'show':0,  'C':0,
          'o':0, 'on':0, 'P':0,
          'mp':0,
          'v':0,
}


#########################################################
###### start main program function

def main(args={}):
#########################################################
############ loading options
  global opt
  if not args: opt=command_line(def_opt, help_msg, 'io', synonyms=command_line_synonyms )
  else:  opt=args
  set_MMlib_var('opt', opt)
  #global temp_folder; temp_folder=Folder(random_folder(opt['temp'])); test_writeable_folder(temp_folder, 'temp_folder'); set_MMlib_var('temp_folder', temp_folder)
  #global split_folder;    split_folder=Folder(opt['temp']);               test_writeable_folder(split_folder); set_MMlib_var('split_folder', split_folder) 
  #checking input
  input_file=opt['i'];   check_file_presence(input_file, 'input_file')
  intreeformat=opt['if'];    
  outtreeformat=intreeformat if opt['of'] is None else opt['of']
  tree=ete3.Tree(input_file, format=intreeformat)

  if opt['names']:
    dict_names={}; n_not_found_tree=0; 
    for line in open(opt['names']): 
      if not line[:-1]: continue
      splt=line.strip().split('\t')
      dict_names[splt[0]]=splt[1]
    for n in tree.traverse():
      if n.name and not n.name in dict_names: n_not_found_tree+=1
      elif n.name:   
        replacement=dict_names[n.name]
        dict_names[n.name]=None        
        n.name= replacement

    #print dict_names
    n_not_found_dict= len( [x for x in dict_names if not dict_names[x] is None ])
    if n_not_found_tree:  printerr('WARNING {n} titles in the tree were not found in {f}'.format(f=opt['names'], n=n_not_found_tree), 1)
    if n_not_found_dict:  printerr('WARNING {n} titles in {f} were not found in the tree'.format(f=opt['names'], n=n_not_found_dict), 1)
    write('-names : node names were modified from {f} '.format(f=opt['names']), 1)

  if opt['mask']:
    for n in tree.traverse():      n.name= mask_characters(n.name).replace(' ', '_')
    write('-mask : masked node names and replaced spaces with underscores', 1)

  if opt['unmask']:
    for n in tree.traverse():      n.name= unmask_characters(   n.name.replace('_', ' ') )
    write('-unmask : unmasked node names and replaced underscores with spaces', 1)

  if opt['exp']:
    repr2members={}
    for line in open(opt['exp']):
      if not line.strip(): continue
      splt=line.strip().split('\t')
      repr2members[splt[0]]=splt[1].split()

    for rep in repr2members:
      rep_node=tree&rep
      new_node=rep_node.up.add_child()
      rep_node.detach()
      new_node.add_child(rep_node)
      sisters_node=new_node.add_child()
      for sister in repr2members[rep]:
        sisters_node.add_child(name=sister)
    write('-exp : expanded nodes adding sisters from {f}'.format(f=opt['exp']), 1)

  if opt['log']:
    for n in tree.traverse():      
      n.dist=log10(1+n.dist)
    write('-log : transformed distances with log10 function', 1)

  if opt['prn']:
    names=[s.strip() for s in open(opt['prn']) if s.strip()]
    tree.prune(names, preserve_branch_length=False)
    write('-prn : pruned tree', 1)

  if opt['std']:
    tree.standardize()
    write('-std : standardized tree structure', 1)

  if opt['lad']:
    tree.ladderize()
    write('-lad : ladderized tree structure', 1)

  if opt['um']:
    tl=float(opt['um'])  # 1.0 if no argument is provided
    if tl == -1: tl=None
    tree.convert_to_ultrametric(tl)
    write('-um : converted to ultrametric, tree length= {tl}'.format(tl=tl), 1)


  if opt['mp']:
    write('-mp  : Set outgroup using midpoint method', 1)
    tree.set_outgroup(tree.get_midpoint_outgroup())

  if opt['lca']:
    lca_nodes=opt['lca'].split(':')
    tree=tree.get_common_ancestor(*lca_nodes)
    write('-lca : Zoom into the last common ancestor of {}'.format(' and '.join(lca_nodes)), 1)

#####

  if opt['print']:
    write('-print : find below the tree in ASCII format:', 1)
    print(tree)

  if opt['show']:
    tree_style=ete3.TreeStyle()
    node_style=ete3.NodeStyle()
    node_style['size']=0
    for n in tree.traverse(): n.set_style(node_style)
    if opt['C']: tree_style.mode='c'
    if opt['show']==1:  
      write('-show : Opening ete interactive environment ...', 1)
      tree.show(tree_style=tree_style)
    else:               
      write('-show : printing tree render to file {f} ... '.format(f=opt['show']), 1)
      tree.render(opt['show'], tree_style=tree_style)

  if opt['o']:
    features=None if not opt['off'] else (
        [] if opt['off']=='all' else opt['off'].split(','))      
    write_to_file(  tree.write(format=outtreeformat,
                               features=features),
                    opt['o'] )
    write('|', 1)
    write('--> Wrote output file {f} in newick format {fr}'.format(f=opt['o'], fr=outtreeformat), 1)

  if opt['on']:
    write_to_file( '\n'.join([n.name for n in tree]), opt['on'] )
    write('|', 1)
    write('--> Wrote output file {f} with leaf names'.format(f=opt['on']), 1)

  if opt['P']:
    import IPython
    IPython.embed(header='Interactive shell! Your tree is in variable -> tree ')
  
  

  

  ###############



#######################################################################################################################################

def close_program():
  if 'temp_folder' in globals() and is_directory(temp_folder):
    bbash('rm -r '+temp_folder)
  try:
    if get_MMlib_var('printed_rchar'): 
      printerr('\r'+printed_rchar*' ' ) #flushing service msg space       
  except:
    pass

  if 'log_file' in globals(): log_file.close()


if __name__ == "__main__":
  try:
    main()
    close_program()  
  except Exception:
    close_program()
    raise 
